"""
Test on graph compression.

"""
import unittest
import tempfile

import powergrasp
from powergrasp import commons


class TestUnambiguousCompression(unittest.TestCase):
    """This class yields many subtests about unambiguous compression results.

    The test case and associated results are defined in setUp() method.
    Bubble outputs are comparated, excluding comments and empty lines.

    The unambiguous compression is important : bubble comparison is sensitive
    to powernodes numbers and related.
    However, it resists to multiple connected components.

    """

    def setUp(self):
        self.test_cases = {
            'double_biclique_unambiguous.lp' : RESULT_DDIAMUN,
            'bipartite.lp'                   : RESULT_BIP,
            'testblocks.lp'                  : RESULT_BLO,
            'perfectfit.lp'                  : RESULT_PFC,
            'cliques.lp'                     : RESULT_CLIQUE,
            'star.lp'                        : RESULT_STAR,
            'concomp.lp'                     : RESULT_CC,
            'ecoli_2896-53.gml'              : RESULT_ECOLI,
            'one_edge.lp'                    : RESULT_ONEDGE,
            'empty.lp'                       : '',
        }


    def unified_bubble(self, bubble_lines):
        """Return the set of comparable lines found in given bubble lines"""
        # filter out comments and blank lines
        return set(line.strip() for line in bubble_lines
                   if not line.startswith('#') and len(line.strip()) > 0)

    def unified_bubble_from_file(self, bubble_file):
        """Return the set of comparable lines found in given bubble file"""
        with open(bubble_file) as fd:
            return self.unified_bubble(fd)

    def assert_equality(self, input_filename, expected_bubble):
        """call assertEqual to compare given bubble and bubble generated by
        compressing given graph."""
        tmp = tempfile.NamedTemporaryFile('w', delete=False)
        tmp.close()
        powergrasp.compress(
            input_filename,
            output_file=tmp.name,
            output_format='bbl',
            loglevel='CRITICAL',  # don't show any output, please
        )
        with self.subTest(filename=input_filename):
            self.assertEqual(
                self.unified_bubble_from_file(tmp.name),
                self.unified_bubble(expected_bubble),
            )


    def test_cases(self):
        for filename, expected in self.test_cases.items():
            self.assert_equality(
                commons.test_case(filename),
                expected.split('\n')
            )


# Expected results of tested cases
RESULT_DDIAMUN = """
NODE\t"d"
NODE\t"i"
NODE\t"b"
NODE\t"a"
NODE\t"q"
NODE\t"f"
NODE\t"c"
NODE\t"m2"
NODE\t"p"
NODE\t"m"
NODE\t"o"
NODE\t"n"
NODE\t"e"
NODE\t"l"
NODE\t"h"
NODE\t"d2"
NODE\t"g"
NODE\t"j"
IN\tPWRN-"a"-4-1\tPWRN-"a"-3-2
IN\t"a"\tPWRN-"a"-1-1
IN\t"d"\tPWRN-"a"-1-1
IN\t"d2"\tPWRN-"a"-1-1
IN\t"m"\tPWRN-"a"-5-1
IN\t"m2"\tPWRN-"a"-5-1
IN\t"l"\tPWRN-"a"-3-1
IN\t"p"\tPWRN-"a"-3-1
IN\t"b"\tPWRN-"a"-1-2
IN\t"o"\tPWRN-"a"-1-2
IN\t"e"\tPWRN-"a"-1-2
IN\t"q"\tPWRN-"a"-1-2
IN\t"c"\tPWRN-"a"-1-2
IN\tPWRN-"a"-3-1\tPWRN-"a"-2-1
IN\t"f"\tPWRN-"a"-2-1
IN\t"g"\tPWRN-"a"-2-1
IN\t"j"\tPWRN-"a"-2-1
IN\t"h"\tPWRN-"a"-2-2
IN\t"i"\tPWRN-"a"-2-2
IN\t"n"\tPWRN-"a"-4-1
IN\tPWRN-"a"-5-1\tPWRN-"a"-4-1
EDGE\tPWRN-"a"-1-1\tPWRN-"a"-1-2\t1.0
EDGE\tPWRN-"a"-5-1\t"n"\t1.0
EDGE\tPWRN-"a"-3-1\tPWRN-"a"-3-2\t1.0
EDGE\t"b"\t"c"\t1.0
EDGE\tPWRN-"a"-2-1\tPWRN-"a"-2-2\t1.0
EDGE\tPWRN-"a"-4-1\t"q"\t1.0
EDGE\t"f"\t"g"\t1.0
"""


RESULT_BLO = """
NODE\t"f"
NODE\t"d"
NODE\t"k"
NODE\t"i"
NODE\t"a"
NODE\t"h"
NODE\t"e"
NODE\t"g"
NODE\t"c"
NODE\t"b"
NODE\t"j"
IN\t"e"\tPWRN-"a"-2-1
IN\tPWRN-"a"-3-1\tPWRN-"a"-2-1
IN\t"a"\tPWRN-"a"-3-1
IN\t"c"\tPWRN-"a"-3-1
IN\t"b"\tPWRN-"a"-3-1
IN\t"d"\tPWRN-"a"-3-1
IN\t"g"\tPWRN-"a"-4-1
IN\t"f"\tPWRN-"a"-4-1
IN\t"h"\tPWRN-"a"-4-1
IN\t"i"\tPWRN-"a"-1-1
IN\t"j"\tPWRN-"a"-1-1
IN\tPWRN-"a"-4-1\tPWRN-"a"-1-1
IN\t"k"\tPWRN-"a"-1-1
EDGE\tPWRN-"a"-2-1\tPWRN-"a"-2-1\t1.0
EDGE\tPWRN-"a"-3-1\t"f"\t1.0
EDGE\tPWRN-"a"-4-1\t"l"\t1.0
EDGE\tPWRN-"a"-1-1\tPWRN-"a"-1-1\t1.0
"""

RESULT_BIP = """
NODE\t"b"
NODE\t"d"
NODE\t"a"
NODE\t"f"
NODE\t"e"
NODE\t"c"
IN\t"e"\tPWRN-"a"-2-1
IN\t"c"\tPWRN-"a"-2-1
IN\t"d"\tPWRN-"a"-2-1
IN\tPWRN-"a"-2-1\tPWRN-"a"-1-2
IN\t"a"\tPWRN-"a"-1-1
IN\t"f"\tPWRN-"a"-1-1
IN\t"b"\tPWRN-"a"-1-1
EDGE\tPWRN-"a"-2-1\tPWRN-"a"-2-1\t1.0
EDGE\tPWRN-"a"-1-1\tPWRN-"a"-1-2\t1.0
"""

RESULT_CLIQUE = """
NODE\t"c"
NODE\t"b"
NODE\t"a"
NODE\t"d"
IN\t"c"\tPWRN-"a"-1-1
IN\t"b"\tPWRN-"a"-1-1
IN\t"a"\tPWRN-"a"-1-1
IN\t"d"\tPWRN-"a"-1-1
EDGE\tPWRN-"a"-1-1\tPWRN-"a"-1-1\t1.0
"""

RESULT_STAR = """
NODE\t"c"
NODE\t"d"
NODE\t"b"
NODE\t"e"
IN\t"c"\tPWRN-"a"-1-2
IN\t"d"\tPWRN-"a"-1-2
IN\t"b"\tPWRN-"a"-1-2
IN\t"e"\tPWRN-"a"-1-2
EDGE\tPWRN-"a"-1-2\t"a"\t1.0
"""

RESULT_CC = """
EDGE\t"23"\t"42"\t1.0
NODE\t"2"
NODE\t"5"
NODE\t"3"
NODE\t"4"
NODE\t"6"
NODE\t"1"
IN\t"1"\tPWRN-"1"-1-1
IN\t"6"\tPWRN-"1"-1-1
IN\t"4"\tPWRN-"1"-1-2
IN\t"2"\tPWRN-"1"-1-2
IN\t"5"\tPWRN-"1"-1-2
IN\t"3"\tPWRN-"1"-1-2
EDGE\tPWRN-"1"-1-1\tPWRN-"1"-1-2\t1.0
NODE\t"b"
NODE\t"c"
NODE\t"d"
NODE\t"e"
IN\t"b"\tPWRN-"a"-1-2
IN\t"c"\tPWRN-"a"-1-2
IN\t"d"\tPWRN-"a"-1-2
IN\t"e"\tPWRN-"a"-1-2
EDGE\tPWRN-"a"-1-2\t"a"\t1.0
"""

RESULT_PFC = """
NODE\t"j"
NODE\t"l"
NODE\t"b"
NODE\t"g"
NODE\t"h"
NODE\t"a"
NODE\t"d"
NODE\t"c"
NODE\t"e"
NODE\t"f"
NODE\t"i"
IN\tPWRN-"a"-2-1\tPWRN-"a"-1-2
IN\t"h"\tPWRN-"a"-1-2
IN\t"i"\tPWRN-"a"-1-2
IN\t"b"\tPWRN-"a"-1-1
IN\tPWRN-"a"-3-1\tPWRN-"a"-1-1
IN\t"a"\tPWRN-"a"-1-1
IN\t"j"\tPWRN-"a"-3-2
IN\t"l"\tPWRN-"a"-3-2
IN\t"d"\tPWRN-"a"-3-1
IN\t"c"\tPWRN-"a"-3-1
IN\t"e"\tPWRN-"a"-2-1
IN\t"g"\tPWRN-"a"-2-1
IN\t"f"\tPWRN-"a"-2-1
IN\tPWRN-"a"-3-2\tPWRN-"a"-2-2
EDGE\tPWRN-"a"-3-1\tPWRN-"a"-3-2\t1.0
EDGE\tPWRN-"a"-2-1\tPWRN-"a"-2-2\t1.0
EDGE\tPWRN-"a"-1-1\tPWRN-"a"-1-2\t1.0
"""

RESULT_ECOLI = """
NODE\t"R_VALabc"
NODE\t"R_DHAD1"
NODE\t"R_VALt2r"
IN\t"R_VALabc"\tPWRN-"R_DHAD1"-1-1
IN\t"R_DHAD1"\tPWRN-"R_DHAD1"-1-1
IN\t"R_VALt2r"\tPWRN-"R_DHAD1"-1-1
EDGE\tPWRN-"R_DHAD1"-1-1\tPWRN-"R_DHAD1"-1-1\t1.0
NODE\t"RXN-14213"
NODE\t"RXN-14200"
NODE\t"R_DTMPK"
IN\t"RXN-14213"\tPWRN-"RXN-14200"-1-1
IN\t"RXN-14200"\tPWRN-"RXN-14200"-1-1
IN\t"R_DTMPK"\tPWRN-"RXN-14200"-1-1
EDGE\tPWRN-"RXN-14200"-1-1\tPWRN-"RXN-14200"-1-1\t1.0
NODE\t"R_EX_26dap_M_LPAREN_e_RPAREN_"
NODE\t"RXN-14246"
NODE\t"R_DAPE"
IN\t"R_EX_26dap_M_LPAREN_e_RPAREN_"\tPWRN-"RXN-14246"-1-1
IN\t"RXN-14246"\tPWRN-"RXN-14246"-1-1
IN\t"R_DAPE"\tPWRN-"RXN-14246"-1-1
EDGE\tPWRN-"RXN-14246"-1-1\tPWRN-"RXN-14246"-1-1\t1.0
EDGE\t"NADH-KINASE-RXN"\t"R_NADK"\t1.0
NODE\t"R_G1PACT"
NODE\t"GLUCOSAMINEPNACETYLTRANS-RXN"
NODE\t"R_PGAMT"
NODE\t"PHOSACETYLGLUCOSAMINEMUT-RXN"
NODE\t"R_AMANAPE"
NODE\t"R_EX_acgam_LPAREN_e_RPAREN_"
IN\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-1
IN\t"PHOSACETYLGLUCOSAMINEMUT-RXN"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-1
IN\t"GLUCOSAMINEPNACETYLTRANS-RXN"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1
IN\t"R_AMANAPE"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1
IN\t"R_EX_acgam_LPAREN_e_RPAREN_"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1
IN\t"R_G1PACT"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-2
IN\t"R_PGAMT"\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-2
EDGE\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-1\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-1-2\t1.0
EDGE\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1\tPWRN-"GLUCOSAMINEPNACETYLTRANS-RXN"-2-1\t1.0
NODE\t"R_GLGC"
NODE\t"RXN-10770"
NODE\t"GLYMALTOPHOSPHORYL-RXN"
IN\t"R_GLGC"\tPWRN-"GLYMALTOPHOSPHORYL-RXN"-1-1
IN\t"RXN-10770"\tPWRN-"GLYMALTOPHOSPHORYL-RXN"-1-1
IN\t"GLYMALTOPHOSPHORYL-RXN"\tPWRN-"GLYMALTOPHOSPHORYL-RXN"-1-1
EDGE\tPWRN-"GLYMALTOPHOSPHORYL-RXN"-1-1\tPWRN-"GLYMALTOPHOSPHORYL-RXN"-1-1\t1.0
"""

RESULT_TESTGML = """
NODE\t"g"
NODE\t"b"
NODE\t"m"
NODE\t"s"
NODE\t"d"
NODE\t"v"
NODE\t"w"
NODE\t"f"
NODE\t"c"
IN\t"d"\tPWRN-"b"-1-1
IN\t"w"\tPWRN-"b"-1-1
IN\t"c"\tPWRN-"b"-1-1
IN\t"s"\tPWRN-"b"-1-1
IN\t"b"\tPWRN-"b"-1-1
IN\t"f"\tPWRN-"b"-2-2
IN\tPWRN-"b"-3-2\tPWRN-"b"-2-2
IN\t"g"\tPWRN-"b"-2-2
IN\t"m"\tPWRN-"b"-3-2
IN\t"v"\tPWRN-"b"-3-2
EDGE\tPWRN-"b"-3-2\t"l"\t1.0
EDGE\t"l"\t"p"\t1.0
EDGE\tPWRN-"b"-1-1\tPWRN-"b"-1-1\t1.0
EDGE\t"m"\t"v"\t1.0
EDGE\t"c"\t"f"\t1.0
EDGE\tPWRN-"b"-2-2\t"b"\t1.0
"""

RESULT_ONEDGE = """
EDGE\t"a"\t"b"\t1.0
"""



from .observer import Signals, Priorities, CompressionObserver
from .counters import ConnectedComponentsCounter, ObjectCounter
from .output import InteractiveCompression, LatticeDrawer, OutputWriter
from .time import TimeCounter, NullTimeCounter



# -*- coding: utf-8 -*-
"""
definition of the SBML input format converter.

The data is used as follow:
    - species are nodes;
    - reactions are nodes;
    - an edge is created between each reaction node and each species involved in it;

"""
import itertools

from powergrasp import commons
from powergrasp.converter.input_converter import InConverter


LOGGER = commons.logger()
NAME_PREFIX = 'metacyc:'


class InSBML(InConverter):
    """Convert given SBML file in ASP file"""
    FORMAT_NAME = 'sbml'
    FORMAT_EXTENSIONS = ('sbml',)

    def _gen_edges(self, filename_sbml:str) -> dict:
        """Yields pair (node, successor), representing the data contained
        in input sbml file.
        """
        try:
            yield from sbml_to_atom_generator(filename_sbml)
        except IOError as e:
            LOGGER.error(self.error_input_file(filename_sbml, e))
        except ImportError:
            LOGGER.error("libsbml module is necessary for use SBML as input"
                         " format. 'pip install libsbml' should do the job."
                         " Compression aborted.")
            exit(1)
        return  # empty generator pattern
        yield


def sbml_to_atom_generator(filename:str) -> dict:
    from libsbml import readSBML

    document = readSBML(filename)
    level    = document.getLevel()
    version  = document.getVersion()
    model    = document.getModel()

    LOGGER.info('libsbml found a SBML data of level ' + str(level)
                + ' and of version ' + str(version) + '.')

    # print lib fatal error of the libsbml
    errors = (document.getError(idx) for idx in itertools.count())
    errors = (err for err in itertools.takewhile(lambda e: e is not None, errors)
              if err.isError() or err.isFatal())
    for error in errors:
        LOGGER.error('libsbml error on input file: ' + error.getMessage().strip())

    if (model == None):
        LOGGER.error('libsbml: No model found in file ' + filename + '.' )
        exit(1)

    # build dictionnary that link species id with its name
    species_name = {}
    for specie in model.getListOfSpecies():
        species_name[specie.getId()] = specie.getName().lstrip(NAME_PREFIX)

    # get reactions, produces all edges in the outputed dict
    for reaction in model.getListOfReactions():
        name = reaction.getName().lstrip(NAME_PREFIX)
        products  = (species_name[p.getSpecies()] for p in reaction.getListOfProducts() )
        reactants = (species_name[p.getSpecies()] for p in reaction.getListOfReactants())
        for node in itertools.chain(products, reactants):
            yield name, node



"""
Implementation of the compression routine.

"""
import time
import itertools
from functools import partial

from powergrasp.commons import ASP_ARG_UPPERBOUND, ASP_ARG_CC
from powergrasp.commons import ASP_ARG_LOWERBOUND, ASP_ARG_STEP

from powergrasp import observers
from powergrasp import solving
from powergrasp import commons
from powergrasp import atoms
from powergrasp import utils

from powergrasp.observers import Signals  # shortcut


LOGGER = commons.logger()
# under this minimal score, the found concept is not interesting
MINIMAL_SCORE = 2


def search_concept(input_atoms, score_min, score_max, cc, step, aspconfig):
    """Return the concept found and its score.

    if no concept found: return (None, None)
    if concept found: return (atoms, concept score)

    """
    if score_min > score_max: return None, None
    search_clique = commons.ASP_SRC_FINDCC in aspconfig.files
    CONCEPT_NAME = ('' if search_clique else 'BI') + 'CLIQUE'
    LOGGER.debug('FIND BEST ' + CONCEPT_NAME
                 + ' [' + str(score_min) + ';' + str(score_max) + ']')
    model = solving.model_from(
        base_atoms=input_atoms,
        aspargs={ASP_ARG_CC: cc, ASP_ARG_STEP: step,
                 ASP_ARG_LOWERBOUND: score_min,
                 ASP_ARG_UPPERBOUND: score_max},
        aspconfig=aspconfig,
    )
    # treatment of the model
    if model is None:
        LOGGER.debug(CONCEPT_NAME + ' SEARCH: no model found')
        concept_score = None
    else:
        assert('score' in str(model))
        concept_score = int(model.get_first('score').arguments[0])
    return model, concept_score


def compress_lp_graph(graph_lp, *, all_observers=[],
                      extract_config=None, biclique_config=None,
                      clique_config=None):
    """apply the compression algorithm on given graph. Yield lines of
    bubble file.

    graph_lp: filename containing the input graph in ASP readable format.
    all_observers: iterable of observers that needs to be updated when
        something happens.
    extract_config: ASP configuration for the extraction.
    biclique_config: ASP configuration for the biclique search.
    clique_config: ASP configuration for the clique search.

    """
    # None to default value
    if extract_config is None:
        extract_config = solving.CONFIG_EXTRACTION()
    if biclique_config is None:
        biclique_config = solving.CONFIG_BICLIQUE_SEARCH()
    if clique_config is None:
        clique_config = solving.CONFIG_CLIQUE_SEARCH()
    # Shortcuts and curried functions
    def notify_observers(*args, **kwargs):
        "Notify observers with given signals"
        for observer in all_observers:
            observer.update(*args, **kwargs)
    search_clique = partial(search_concept, aspconfig=clique_config)
    search_biclique = partial(search_concept, aspconfig=biclique_config)

    # INIT
    # Extract graph data
    LOGGER.info('#################')
    LOGGER.info('#### EXTRACT ####')
    LOGGER.info('#################')
    notify_observers(
        Signals.CompressionStarted,
        Signals.ExtractionStarted,
        asp_config_updated=(extract_config, clique_config, biclique_config)
    )
    # creat a solver that get all information about the graph
    connected_components = (solving.all_models_from(
        '', aspfiles=[graph_lp], aspconfig=extract_config,
    ))
    notify_observers(Signals.ExtractionStopped)

    # ITERATIVE TREATMENT
    # Find connected components
    LOGGER.info('#################')
    LOGGER.info('####   CC    ####')
    LOGGER.info('#################')
    total_edges_counter = 0  # number of edge in all the graph (for statistics)
    total_remain_edges_counter = 0
    for cc_nb, cc_atoms in enumerate(connected_components):
        # get data from cc_atoms
        cc_name = cc_atoms.get_first('cc').arguments[0]
        notify_observers(cc_count_generated=int(cc_atoms.get_first('nb_cc').arguments[0]))
        assert any(isinstance(cc_name, cls) for cls in (str, int))
        remain_edges_cc = tuple(atom for atom in cc_atoms
                                if atom.predicate == 'oedge')
        def nb_cc_edges(): return len(remain_edges_cc)
        total_edges_counter += nb_cc_edges()
        previous_atoms = atoms.to_str(cc_atoms)
        cc_atoms = atoms.to_str(cc_atoms)
        # treatment of the connected_components stated
        notify_observers(connected_component_started=(cc_nb, cc_name))
        # contains interesting atoms and the non covered edges at last step
        model_found_at_last_iteration = True  # False when no model found
        # main loop
        step = 0
        last_score = nb_cc_edges()  # score of the previous step, or maximal score
        # iteration
        notify_observers(Signals.IterationStarted)
        while model_found_at_last_iteration:
            # STEP INITIALIZATION
            notify_observers(Signals.StepStarted)
            step += 1
            concept_score, best_model = None, None
            score_lowerbound = MINIMAL_SCORE

            clique_model, clique_score = search_clique(
                previous_atoms,
                score_min=score_lowerbound,
                score_max=last_score,
                cc=cc_name,
                step=step,
            )

            if not clique_model or score_lowerbound <= clique_score:
                model, score = search_biclique(
                    previous_atoms,
                    score_min=clique_score if clique_score else score_lowerbound,
                    score_max=last_score,
                    cc=cc_name,
                    step=step,
                )
            best_model = model if model else clique_model
            best_score = score if score else clique_score

            # best_model is computed
            last_score = best_score


            #########################
            LOGGER.debug('BEST MODEL TREATMENT\n\n\n')
            #########################
            # stop cc compression if no model found
            if best_model is None:
                # Send signal with no data ; this is not necessary, but allow
                # the plotting to print all the iterations. Without that signal,
                # the final result cannot show one iteration per connected
                # component, while for each of them the last step will not
                # produce any model. Next line handle the model inexistence.
                notify_observers(step_data_generated=[None] * 3)
                model_found_at_last_iteration = False
            else:  # at least one model was found, and the best is best_model
                # debug printing
                LOGGER.debug('POWERNODES:\n\t' + atoms.prettified(
                    best_model,
                    names=('powernode', 'poweredge', 'score'),
                    joiner='\n\t',
                    sort=True
                ))
                # new concept is the previous concept of next concept
                assert(isinstance(best_score, int))
                assert(best_score >= MINIMAL_SCORE)
                last_score = best_score
                remain_edges_cc = tuple(best_model.get('oedge'))
                previous_atoms = atoms.to_str(best_model.get(
                    ('oedge', 'membercc', 'block', 'include_block')
                ))

                # give new powernodes to converter
                notify_observers(model_found=tuple(
                    atom for atom in best_model
                    if atom.predicate in ('powernode', 'clique', 'poweredge')
                ))

                final_concept = tuple(atom for atom in best_model
                                      if atom.predicate in ('powernode',
                                                            'poweredge'))

                # save the number of generated powernodes and poweredges
                new_powernode_count = next(
                    int(atom.arguments[0]) for atom in best_model
                    if atom.predicate == 'powernode_count')
                new_poweredge_count = next(
                    int(atom.arguments[0]) for atom in best_model
                    if atom.predicate == 'poweredge_count')
                if new_powernode_count not in (0, 1, 2):
                    LOGGER.error('Error of Powernode generation: '
                                 + str(new_powernode_count) + 'generated.'
                                 + ('It can be a problem of stars that are counted as powernodes'
                                    if new_powernode_count < 0 else
                                    'Too many powernodes for one step.')
                                 + ' It\'s probable that this problem will only'
                                 + ' touch the statistics, but the compression'
                                 + ' itself will not be compromised.'
                                )

                # notify_observers: provide the data
                notify_observers(
                    model_found=final_concept,
                    step_data_generated=(new_powernode_count,
                                         new_poweredge_count, best_score)
                )
                if nb_cc_edges() < 0:
                    print('REMAIN_EDGES_GLOBAL:', nb_cc_edges())
                    assert(nb_cc_edges() >= 0)
            # notify_observers:
            notify_observers(Signals.StepStopped)
            notify_observers(Signals.StepFinalized)

        notify_observers(Signals.IterationStopped)
        # END while model_found_at_last_iteration
        # Here, all models was processed in the connected component

        # Management of remain data in the connected component
        assert(nb_cc_edges() >= 0)

        # Remain edges in cc
        notify_observers(cc_remain_edge_generated=remain_edges_cc)
        total_remain_edges_counter += nb_cc_edges()


        notify_observers(Signals.ConnectedComponentStopped)

    # DEINIT
    LOGGER.info('#################')
    LOGGER.info('#### RESULTS ####')
    LOGGER.info('#################')
    # compute a human readable final results string,
    # and put it in the output and in level info.
    notify_observers(final_edge_count_generated=total_edges_counter,
                     final_remain_edge_count_generated=total_remain_edges_counter)
    notify_observers(Signals.CompressionStopped)
    notify_observers(Signals.CompressionFinalized)



"""
Definition of various output-related compression observers.

"""
import sys

from .observer  import CompressionObserver, Signals, Priorities
from powergrasp import commons
from powergrasp import converter
from powergrasp import utils


LOGGER = commons.logger()


class InteractiveCompression(CompressionObserver):
    """Make the compression interactive: ask for an input from user
    at the end of each step"""
    def _update(self, signals):
        if Signals.StepStopped in signals:
            try:
                input('<hit enter for next model computation>')
            except KeyboardInterrupt:
                exit()

    @property
    def priority(self):
        return Priorities.Minimal


class LatticeDrawer(CompressionObserver):
    """Draw a lattice of all connected component"""

    def __init__(self, directory=commons.PACKAGE_DIR_DATA,
                 prefix='lattice_'):
        self.basename = directory + prefix + '{}'
        self.last_cc = None

    def _update(self, signals):
        if Signals.ConnectedComponentStarted in signals:
            _, self.last_cc = signals[Signals.ConnectedComponentStarted]
        if self.last_cc is not None and Signals.PreprocessingStopped in signals:
            atoms = signals[Signals.PreprocessingStopped]
            graphdict = utils.asp2graph(atoms)
            cc_filename = self.basename.format(self.last_cc)
            utils.line_diagram(graphdict, filename=cc_filename)
            LOGGER.info('Line diagram of CC ' + str(self.last_cc)
                        + ' saved in ' + cc_filename)
            self.last_cc = None  # don't react next iteration of the same connected component


class OutputWriter(CompressionObserver):
    """Manage the output file, and conversion into expected output format"""

    def __init__(self, output_filename, output_format):
        format = OutputWriter.format_deduced_from(output_filename, output_format)
        self.output = open(output_filename, 'w') if output_filename else sys.stdout
        self.converter = converter.output_converter_for(format)

    def _update(self, signals):
        # print('Output Writer:', signals)
        if Signals.ModelFound in signals:
            # give new powernodes to converter
            atoms = signals[Signals.ModelFound]
            self.converter.convert(atoms)
        if Signals.ConnectedComponentStopped in signals:
            self.output.write(self.converter.finalized())
            self.converter.reset_containers()
            LOGGER.debug('Final data saved in file ' + self.output.name)
        if Signals.CCRemainEdgeGenerated in signals:
            remain_edges = signals[Signals.CCRemainEdgeGenerated]
            self.converter.convert(remain_edges)
        if Signals.CompressionStopped in signals:
            if self.output is not sys.stdout: self.output.close()
        if Signals.CompressionStarted in signals:
            self.output.write(self.converter.header())

    def comment(self, lines):
        """Add given lines to output as comments"""
        self.output.write(self.converter.comment(lines))

    @property
    def priority(self):
        return Priorities.Minimal

    @staticmethod
    def format_deduced_from(output_file, output_format):
        """Return the most likely expected output format by looking at given args"""
        # look at the output_file extension if output_format is unusable
        if not output_format or output_format not in converter.OUTPUT_FORMATS:
            try:
                output_format = output_file.split('.')[-1]  # extension of the file
            except (IndexError, AttributeError):
                output_format = converter.DEFAULT_OUTPUT_FORMAT  # use BBL
        # verifications
        if output_format not in converter.OUTPUT_FORMATS:
            output_format = converter.DEFAULT_OUTPUT_FORMAT
        return output_format



"""
Definitions of observers related to the time.

"""
import time
from functools import partial

from .observer import CompressionObserver, Signals, Priorities
from .observer import SIGNAL_STARTED, SIGNAL_STOPPED
from powergrasp import commons


LOGGER = commons.logger()


class TimeCounter(CompressionObserver):
    """Observer that looks for all received signals, and creates a timer for
    any signal with the value finishing by '_started', and destroys the timer
    after logging its value when receiving the sibling '_stopped' signal.

    This allows a generalist and automatic benchmarking
    of the observed phenomenoms.

    """
    TIMER_PREFIX = '_timer_'  # used for generated attributes
    LAST_PREFIX = '_last_time_'  # used for generated attributes

    def __init__(self, round_ndigits=3, log_times=True, ignore=[]):
        """Wait for the ndigits parameter of the builtin round function.
        If log_times is False, no time will be logged.
        Ignore must be an iterable of Signals *Started to ignore.

        """
        self.log_times = log_times
        if round_ndigits:
            self.round = partial(round, ndigits=round_ndigits)
        else:
            self.round = round
        self.ignored = set(ignore)
        self.ignored |= set(  # add stopped signals
            Signals(s.value[:-len(SIGNAL_STARTED)] + SIGNAL_STOPPED)
            for s in self.ignored
        )
        assert all(s in Signals for s in self.ignored)


    def _update(self, signals):
        for signal in signals:
            if signal in self.ignored: continue
            name = signal.value
            if name.endswith(SIGNAL_STARTED):
                prop = TimeCounter.property_named_from(name)
                if hasattr(self, prop):
                    self.log_timer(prop)
                setattr(self, prop, time.time())
            if name.endswith(SIGNAL_STOPPED):
                # stop the timer, log its value and save it
                prop = TimeCounter.property_named_from(name)
                setattr(self, prop, time.time() - getattr(self, prop))
                if hasattr(self, prop):
                    self.log_timer(prop)
                    prop_save = TimeCounter.property_named_from(name, prefix=TimeCounter.LAST_PREFIX)
                    setattr(self, prop_save, getattr(self, prop))
                    delattr(self, prop)
                else:
                    LOGGER.warning('TimeCounter: the process '
                                   + TimeCounter.prettified(prop)
                                   + ' is stopped but not started.')

    def __get_property_or_none(self, prop):
        """Return given property value for self (use getattr), or None
        in case of AttributeError"""
        try:
            return getattr(self, prop)
        except AttributeError:
            return None

    @property
    def compression_time(self):
        prop = TimeCounter.property_named_from(Signals.CompressionStarted.value,
                                               prefix=TimeCounter.LAST_PREFIX)
        return self.__get_property_or_none(prop)

    @property
    def extraction_time(self):
        prop = TimeCounter.property_named_from(Signals.ExtractionStarted.value,
                                               prefix=TimeCounter.LAST_PREFIX)
        return self.__get_property_or_none(prop)

    @property
    def last_step_time(self):
        prop = TimeCounter.property_named_from(Signals.StepStarted.value,
                                               prefix=TimeCounter.LAST_PREFIX)
        return self.__get_property_or_none(prop)

    def log_timer(self, property_name):
        "Perform the logging output for the given property"
        timer = self.round(getattr(self, property_name))
        if self.log_times:
            LOGGER.info(self.__class__.__name__ + ': ' + str(timer)
                        + 's for ' + TimeCounter.prettified(property_name) + '.')

    @staticmethod
    def property_named_from(signal_name, prefix=TIMER_PREFIX):
        "Return the property name deduced from the signal name"
        assert len(SIGNAL_STARTED) == len(SIGNAL_STOPPED)
        return prefix + signal_name[:-len(SIGNAL_STARTED)]

    @staticmethod
    def prettified(property_name):
        "Return the prettified property name deduced from the property name"
        property_name = property_name[len(TimeCounter.TIMER_PREFIX):]
        return property_name.replace('_', ' ').title()

    @property
    def priority(self):
        return Priorities.Maximal


class NullTimeCounter(CompressionObserver):
    """Null Pattern applicated to the TimeCounter class"""

    def __init__(self, round_ndigits=3, log_times=True, ignore=[]):
        pass

    def _update(self, signals):
        pass

    @property
    def compression_time(self):
        return 0.

    @property
    def extraction_time(self):
        return 0.

    @property
    def last_step_time(self):
        return 0.

    def log_timer(self, property_name):
        pass

    @staticmethod
    def property_named_from(signal_name, prefix=''):
        pass

    @staticmethod
    def prettified(property_name):
        return ''

    @property
    def priority(self):
        return Priorities.Maximal



"""
usage:
    __main__.py [options]

options:
    --help, -h
    --version, -v
    --graph-data=FILE       filepath to input graph definition
    --output-file=NAME      output file or dir
    --output-format=NAME    output format (see below for formats)
    --interactive           program ask user for next step
    --show-pre              print preprocessed data in stdout
    --count-model           prints models count in stdout
    --count-cc              prints connected component count in stdout
    --timers                prints times of execution in stdout
    --loglevel=NAME         defines terminal log level
    --logfile=FILE          file where all logs will be put in
    --stats-file=FILE       save csv statistics in FILE
    --plot-stats            plot stats found in stats-file if exist
    --plot-file=FILE        instead of show it, save plot in png FILE
    --profiling             print graph info before compress it
    --thread=INT            use INT thread for ASP solving (0 for automatic detection)
    --draw_lattice=DIR      draw in DIR the concept lattice foreach con. comp.

output formats:
    BBL         formated in Bubble format, readable by CyOog plugin of Cytoscape

input formats:
    ASP         edge/2 atoms, where edge(X,Y) describes a link between X and Y.
    SBML        SBML file. Species and reactions will be translated as nodes.
    GML         Graph Modeling Language file.

"""

import os

from powergrasp.powergrasp import compress
from powergrasp.converter  import OUTPUT_FORMATS
from powergrasp import statistics
from powergrasp import converter
from powergrasp import commons
from powergrasp import plots
from powergrasp import utils
from powergrasp import info


if __name__ == '__main__':
    LOGGER = commons.logger()

    # get options, including CLI
    options = commons.options(cli_doc=__doc__)

    # parse them
    # output format verification
    assert(options['output_format'] in OUTPUT_FORMATS)

    # launch compression
    if options['graph_data']:

        if options['profiling']:
            print(utils.test_integrity(options['graph_data']))

        if options['output_file'] is None:
            pass
        elif os.path.isdir(options['output_file']):
            LOGGER.info('Given output file is not a file, but a directory ('
                        + options['output_file'] + ').'
                        + ' Output file will be placed in it, with the name '
                        + commons.basename(options['graph_data']))
            options['output_file'] = os.path.join(
                options['output_file'],
                commons.basename(options['graph_data'])
            )

        # compression itself
        compress(
            graph_data      = options['graph_data'     ],
            output_file     = options['output_file'    ],
            output_format   = options['output_format'  ],
            interactive     = options['interactive'    ],
            count_model     = options['count_model'    ],
            count_cc        = options['count_cc'       ],
            timers          = options['timers'         ],
            stats_file      = options['stats_file'     ],
            logfile         = options['logfile'        ],
            loglevel        = options['loglevel'       ],
            thread          = options['thread'         ],
            draw_lattice    = options['draw_lattice'   ],
        )

    # plotting if statistics csv file given, and showing or saving requested
    if((options['plot_stats'] or options['plot_file']) and
            options['stats_file']):
        plots(
            options['stats_file'],
            savefile=options['plot_file']
        )





"""
definition of the ASP input format converter.

"""
import itertools

from powergrasp.converter.input_converter import InConverter
import powergrasp.commons as commons
from powergrasp import solving
from powergrasp import atoms

LOGGER = commons.logger()


class InASP(InConverter):
    """Convert given GML file in uniformized data"""
    FORMAT_NAME = 'asp'
    FORMAT_EXTENSIONS = ('lp',)

    def _gen_edges(self, filename):
        """Yields pair (node, successor), representing the data contained
        in input ASP file.
        """
        models = solving.model_from('', filename)
        try:
            for atom in models:
                if atom.predicate == 'edge' and atom.nb_args() == 2:
                    node, succ = atom.arguments
                    yield node, succ
        except IOError as e:
            LOGGER.error(self.error_input_file(filename_sbml, e))
        except TypeError:  # file is probably empty… so do the graph
            return
            yield



"""
Definitions of observers signals and other related data.

"""
from enum import Enum

from powergrasp import commons
from powergrasp import utils


LOGGER = commons.logger()
SIGNAL_STARTED   = '_started'
SIGNAL_STOPPED   = '_stopped'
SIGNAL_GENERATED = '_generated'
SIGNAL_FOUND     = '_found'


class Signals(Enum):
    # Steps signals
    StepStarted               = 'step_started'
    StepStopped               = 'step_stopped'
    CompressionStarted        = 'compression_started'
    CompressionStopped        = 'compression_stopped'
    ConnectedComponentStarted = 'connected_component_started'
    ConnectedComponentStopped = 'connected_component_stopped'
    IterationStarted          = 'iteration_started'  # send before all iterations
    IterationStopped          = 'iteration_stopped'  # send at the end of all iterations
    ExtractionStarted         = 'extraction_started'
    ExtractionStopped         = 'extraction_stopped'
    PreprocessingStarted      = 'preprocessing_started'
    PreprocessingStopped      = 'preprocessing_stopped'
    # Finalizations
    StepFinalized             = 'step_finalized'         # send after the StepStopped signal
    CompressionFinalized      = 'compression_finalized'  # send after the CompressionStopped signal
    # Objects signals
    ModelFound                = 'model_found'
    BicliqueFound             = 'biclique_found'
    CliqueFound               = 'clique_found'
    # Data signals
    CompressionTimeGenerated  = 'compression_time_generated'
    FinalEdgeCountGenerated   = 'final_edge_count_generated'
    FinalRemainEdgeCountGenerated = 'final_remain_edge_count_generated'
    CCRemainEdgeGenerated     = 'cc_remain_edge_generated'
    CCCountGenerated          = 'cc_count_generated'
    StepDataGenerated         = 'step_data_generated'
    ASPConfigUpdated          = 'asp_config_updated'


# Priorities: observers are called with respect to their priority (smaller is after)
class Priorities(Enum):
    Maximal = 100
    High    =  80
    Medium  =  50
    Small   =  20
    Minimal =   0


class CompressionObserver:
    """Base class for all compression observers, where the update method used
    by compression is implemented.

    Derived classes must implements the _update method.
    They can also overwrite the priority method, for ensure to be called before
    others observers.

    """
    def update(self, *args, **kwargs):
        # integrate signals of args in kwargs, convert all into Signals object
        # and finally launch the update
        kwargs.update({sig: None for sig in args})
        kwargs = {Signals(sig): value for sig, value in kwargs.items()}
        assert all(sig in Signals for sig in kwargs.keys())
        self._update(kwargs)

    @property
    def priority(self):
        return Priorities.Medium



# -*- coding: utf-8 -*-
"""
Definitions of many functions that works on compressions statistics.

Statistics are mainly a dictionnary with statistics information inside.

Information is itself a dict that links different parameters
 (number of initial edge, number of power{node,edge},…)
 to the value used for compute statistics.

Statistics information can be printed or saved in files
 in raw or latex tabular format.

"""
import csv

from powergrasp import observers
from powergrasp import commons
from powergrasp.observers import Signals  # shortcut


LOGGER = commons.logger()


# Compression data
# class CompressionData(Enum):
# numbers
INIT_EDGE = 'initial_edge_count'     # number of edges in the input graph
FINL_EDGE = 'remain_edge_count'      # number of edges not compressed since the compression beginning
GENR_PWED = 'poweredge_count'        # number of poweredges created since the beginning
GENR_PWND = 'powernode_count'        # number of powernodes created since the beginning
COMP_EDGE = 'compressed_edge_count'  # number of compressed edge since the beginning
CONV_RATE = 'conversion_rate'        # computed conversion rate
EDGE_RDCT = 'edge_reduction'         # computed edge reduction
COMP_RTIO = 'compression_ratio'      # computed compression ratio
GENR_TIME = 'gentime'                # generation time for last iteration
# general
NETW_NAME = 'network_name'     # name of the input graph (name of the file containing input data)
FILE_DESC = 'file_descriptor'  # file descriptor of the output CSV file, or None if no output expected
FILE_WRTR = 'file_writer'      # CSV writer on the output CSV file, or None if no output expected

# All fields that are useful at the end of the compression are put here
PRINTABLE_FIELD = (
    INIT_EDGE, GENR_PWED, GENR_PWND, CONV_RATE,
    EDGE_RDCT, COMP_RTIO, FINL_EDGE, COMP_EDGE
)


# Output formats
FORMAT_TEX = 'tex'
FORMAT_RAW = 'txt'

# Data for plotting
MEASURES = (GENR_TIME, COMP_EDGE, GENR_PWED, GENR_PWND)
COLORS   = ('black', 'green', 'blue', 'red')
LABELS   = (
    'time per step',
    'remaining edges',
    'generated poweredge',
    'generated powernode',
)


class DataExtractor(observers.CompressionObserver, dict):
    """
    DataExtractor is a repository of various data, used for statistical analysis
     of the compression and other treatments.

    """
    GENERATED_PREFIX = '_generated_'

    def __init__(self, statistics_filename=None, network_name='network',
                 output_converter=None, time_counter=None):
        """The statistics_filename, if not None, must be a valid name of file
        that will be overrided, and will contains some data in csv format.

        The output converter, if given, is used at the end for add comments
         containing the final statistic report in the output file.
        Moreover, if the time_counter object is given, the final statistics
         will contains information about compression time.

        """
        statistics_file = None
        if statistics_filename:
            try:
                statistics_file = open(statistics_filename, 'w')
                statistics_writer = csv.DictWriter(statistics_file,
                                                   fieldnames=MEASURES)
                statistics_writer.writeheader()
            except IOError as e:
                LOGGER.warning("The file "
                               + statistics_filename + " can't be opened."
                               + " No statistics will be saved.")
        else:
            statistics_writer = None
        self.output_converter = output_converter
        self.time_counter = time_counter

        dict.__init__(self, {
            NETW_NAME: network_name,
            INIT_EDGE: 0,
            FINL_EDGE: 0,
            GENR_PWED: 0,
            GENR_PWND: 0,
            GENR_TIME: 0,
            COMP_EDGE: 0,
            FILE_DESC: statistics_file,
            FILE_WRTR: statistics_writer,
        })

    @property
    def compression_time(self):
        if self.time_counter:
            return self.time_counter.compression_time
        return 0.

    @property
    def extraction_time(self):
        if self.time_counter:
            return self.time_counter.extraction_time
        return 0.

    def prettified_configs(self):
        """Yield lines of a prettified view of solvers config"""
        for name, config in self.configs:
            yield from (
                name + ' configuration: ',
                '\tfiles   : ' + ', '.join(commons.basename(_) for _ in config.files),
                '\tgrounder: ' + config.gringo_options,
                '\tsolver  : ' + config.clasp_options,
            )

    def _update(self, signals):
        if Signals.CompressionStopped in signals:
            # create the final result render
            final_results = (
                "All cc have been performed"
                + ((' in ' + str(round(self.compression_time, 3)) + 's.')
                   if self.compression_time else '.')
                + ((' (including extraction in ' + str(round(self.extraction_time, 3)) + ')')
                   if self.extraction_time else '')
                + '\n' + '\n'.join(self.prettified_configs())
                + "\nNow, statistics on "
                + self.stats_output()
            )
            LOGGER.info(final_results)
            if self.output_converter:
                self.output_converter.comment(final_results.split('\n'))
            self.finalize()

        if Signals.ASPConfigUpdated in signals:
            extract, clique, biclique = signals[Signals.ASPConfigUpdated]
            self.configs = (
                ('Extraction', extract),
                ('Find best clique', clique),
                ('Find best biclique', biclique),
            )
        if Signals.FinalEdgeCountGenerated in signals:
            self[INIT_EDGE] = int(signals[Signals.FinalEdgeCountGenerated])
        if Signals.FinalRemainEdgeCountGenerated in signals:
            self[FINL_EDGE] = int(signals[Signals.FinalRemainEdgeCountGenerated])
        if Signals.StepDataGenerated in signals:
            (powernode_count,
             poweredge_count,
             compressed_edge_count) = signals[Signals.StepDataGenerated]
            # defense against a no-data case
            if powernode_count is None:
                assert poweredge_count is None
                assert compressed_edge_count is None
            else:  # all data is given
                self[GENR_PWED] += int(poweredge_count)
                self[GENR_PWND] += int(powernode_count)
                self[COMP_EDGE] += int(compressed_edge_count)
        if Signals.StepStopped in signals:
            if self.time_counter:
                gentime = self.time_counter.last_step_time
            else:
                gentime = 0.
            self.write_csv_data(self[GENR_PWED], self[GENR_PWND],
                                gentime, self[COMP_EDGE])


    def stats_output(self, format=FORMAT_RAW):
        """Return self as string in given format.

        format can be one of available format. Default is raw text.
        """
        assert self[INIT_EDGE] is not None
        assert self[FINL_EDGE] is not None
        assert self[GENR_PWND] is not None
        assert self[GENR_PWED] is not None
        self[CONV_RATE] = conversion_rate(
            self[INIT_EDGE], self[FINL_EDGE], self[GENR_PWED], self[GENR_PWND]
        )
        self[EDGE_RDCT] = edge_reduction(
            self[INIT_EDGE], self[FINL_EDGE], self[GENR_PWED]
        )
        self[COMP_RTIO] = compression_ratio(
            self[INIT_EDGE], self[FINL_EDGE], self[GENR_PWED]
        )
        return _formatted(dict(self), format)


    def finalize(self):
        """Close files"""
        if self[FILE_DESC]:
            self[FILE_WRTR] = None
            self[FILE_DESC].close()


    def save(self, filename, format=FORMAT_TEX, erase=False):
        """Saves itself in given filename in given format.

        format can be one of available format. Default is tabular in tex.
        if erase is True, existing content of filename will be erased.
        """
        # produce and save data
        with open(filename, ('w' if erase else 'a')) as fd:
            f.write(self.output(format))


    def write_csv_data(self, poweredge_count, powernode_count,
                       gentime, compressed_edge_count):
        """Write data in the csv file, if exists, else do nothing"""
        if self[FILE_DESC]:
            self[FILE_WRTR].writerow({
                GENR_TIME: gentime,
                GENR_PWED: poweredge_count,
                GENR_PWND: powernode_count,
                COMP_EDGE: compressed_edge_count,
            })


def _formatted(data, format):
    """Return a string formatted in given format that represent the given data.

    """
    fdata = ''
    if format == FORMAT_RAW:
        fdata = data[NETW_NAME] + ':'
        for field in PRINTABLE_FIELD:
            fdata += '\n\t' + field + ': ' + str(data[field])
    elif format == FORMAT_TEX:
        header  = 'network'
        payload = data[NETW_NAME]
        for field in PRINTABLE_FIELD:
            header  += ' & ' + field
            payload += ' & ' + str(data[field])
        header  += '\\\\'
        payload += '\\\\'
        fdata = header + '\n' + payload
    else:
        raise ValueError('unrecognized format')
    return fdata


# DATA COMPUTATION
def conversion_rate(initial_edge, final_edge, poweredge, powernode):
    """Compute conversion rate"""
    try:
        edge = initial_edge
        poweredge = final_edge + poweredge
        return (edge - poweredge) / powernode
    except ZeroDivisionError:
        return 1.

def edge_reduction(initial_edge, final_edge, poweredge):
    """Compute edge reduction (percentage)"""
    try:
        edge = initial_edge
        poweredge = final_edge + poweredge
        return ((edge - poweredge) / edge) * 100
    except ZeroDivisionError:
        return 100.

def compression_ratio(initial_edge, final_edge, poweredge):
    """Compute data compression ratio"""
    try:
        return initial_edge / (final_edge + poweredge)
    except ZeroDivisionError:
        return 1.



"""
Test on graph compression, using statistical results.

"""
import unittest
import tempfile
import inspect
from functools import partial

import powergrasp
from powergrasp import commons
from powergrasp import observers
from powergrasp.statistics import (INIT_EDGE, FINL_EDGE, GENR_PWED, GENR_PWND,
                                   CONV_RATE, EDGE_RDCT, COMP_RTIO, DataExtractor)


STATS_FIELDS = (INIT_EDGE, FINL_EDGE, GENR_PWED, GENR_PWND,
                CONV_RATE, EDGE_RDCT, COMP_RTIO)
STATS_ROUND = partial(round, ndigits=2)


def compression_results(init_edge, finl_edge, genr_pwed, genr_pwnd,
                        conv_rate, edge_rdct, comp_rtio):
    """Return a dict {statistic key: value} where statistics keys are the seven
    constants imported from powergrasp.statistics module, and where the value
    are those given as parameters.
    """
    _, _, _, args = inspect.getargvalues(inspect.currentframe())
    return {
        globals()[name.upper()]: value
        for name, value in args.items()
    }


# Expected results of tested cases
RESULT_DDIAM = compression_results(
    31, 3, 4, 7, 3.4285714285714284, 77.41935483870968, 4.428571428571429,
)
RESULT_CC = compression_results(
    13, 1, 2, 3, 3.3333333333333335, 76.92307692307693, 4.333333333333333,
)
RESULT_CCOLI = compression_results(
    24, 1, 6, 7, 2.4285714285714284, 70.83333333333334, 3.4285714285714284
)
RESULT_EMPTY = compression_results(
    0, 0, 0, 0, 1.0, 100, 1.0,
)


class TestStatisticalCompression(unittest.TestCase):
    """This class yields many subtests about compression results.

    The test case and associated results are defined in setUp() method.
    Tests are robust to ambiguous compression, as long as paths all leads
    to same results.

    These tests operate only on statistical data, thus can't detect problems
    of compression that don't touch the analysed data.

    Tested parameters are:
        Number of model, connected component, biclique,
            clique, poweredge, powernode and initial/remaining edge.
        Conversion rate, edge reduction and compression ratio.


    """

    def setUp(self):
        self.test_cases = {
            'double_biclique.lp' : RESULT_DDIAM,
            'concomp.lp'         : RESULT_CC,
            'empty.lp'           : RESULT_EMPTY,
            'ecoli_2896-53.gml'  : RESULT_CCOLI,
        }

    def filtered_extracted_stats(self, stats):
        return {
            field: stats.get(field, None) for field in STATS_FIELDS
        }

    def assert_equality(self, input_filename, expected_stats):
        """call assertEqual to compare given bubble and bubble generated by
        compressing given graph."""
        statistics = DataExtractor(network_name=input_filename)
        tmp = tempfile.NamedTemporaryFile('w')

        powergrasp.compress(
            input_filename,
            output_file=tmp.name,
            output_format='bbl',
            loglevel='CRITICAL',  # don't show any output, please
            instanciated_observers=[statistics],
        )
        found_stats = self.filtered_extracted_stats(statistics)
        results = ((field, found_stats[field], expected_stats[field])
                   for field in STATS_FIELDS)
        for field, found, expected in results:
            if isinstance(expected, float):
                expected = STATS_ROUND(expected)
            if isinstance(found, float):
                found = STATS_ROUND(found)
            with self.subTest(filename=input_filename, field=field):
                self.assertAlmostEqual(
                    found,
                    expected,
                )
        tmp.close()

    def test_cases(self):
        for filename, expected in self.test_cases.items():
            self.assert_equality(
                commons.test_case(filename),
                expected
            )





"""
Definition of numerous constants, for paths, names, arguments for ASP solvers.
Moreover, some generalist functions are defined,
 notabily for solving management and logging.

"""

# IMPORTS
import logging, logging.config
import multiprocessing  # get number of available CPU
import pkg_resources  # packaging facilies
import sys
import os

from functools   import partial
from collections import ChainMap

from docopt import docopt

from powergrasp.info import PACKAGE_NAME, PACKAGE_VERSION


# Constants
ASP_FILE_EXTENSION = '.lp'
BUBBLE_FORMAT_ID   = 'bbl'

# Directories (starting from the package level)
DIR_SOURCES     = ''  # sources are inside the package
DIR_DATA        = 'data/'
DIR_TEST_CASES  = 'tests/'
DIR_LOGS        = 'logs/'
DIR_ASP_SOURCES = 'ASPsources/'

# Packaging: access to the file, inside the package,
#   independently of the installation directory
access_packaged_file = partial(pkg_resources.resource_filename, PACKAGE_NAME)

# LOGGING VALUES
LOGGER_NAME       = PACKAGE_NAME
DEFAULT_LOG_FILE  = access_packaged_file(DIR_LOGS + LOGGER_NAME + '.log')
DEFAULT_LOG_LEVEL = logging.DEBUG

# PATH INSIDE PACKAGE
PACKAGE_DIR_DATA = access_packaged_file(DIR_DATA)
PACKAGE_DIR_TESTS = access_packaged_file(DIR_TEST_CASES)

# Optimization values
OPT_LOWERBOUND_CUTOFF = 2  # minimal value for the lowerbound optimization

# ASP SOURCES
def __asp_file(name):
    "path to given asp source file name"
    return access_packaged_file(DIR_ASP_SOURCES + name + ASP_FILE_EXTENSION)
ASP_SRC_EXTRACT   = __asp_file('extract')
ASP_SRC_PREPRO    = __asp_file('preprocessing')
ASP_SRC_FINDCC    = __asp_file('findbestclique')
ASP_SRC_FINDBC    = __asp_file('findbestbiclique')
ASP_SRC_POSTPRO   = __asp_file('postprocessing')
ASP_SRC_INCLUSION = __asp_file('inclusion')

# Constants involved in ASP solving
ASP_ARG_CC   = 'cc'
ASP_ARG_STEP = 'k'
ASP_ARG_UPPERBOUND = 'upperbound'
ASP_ARG_LOWERBOUND = 'lowerbound'

# ASP Solver options
# ASP_GRINGO_OPTIONS = ''  # no default options
# ASP_CLASP_OPTIONS  = ''  # options of solving heuristics
# ASP_CLASP_OPTIONS += ' --configuration=frumpy'
# ASP_CLASP_OPTIONS += ' --heuristic=Vsids'
# ASP_CLASP_OPTIONS += ' -n 0'
# ASP_CLASP_OPTIONS.append('--opt-mode=opt')
# ASP_CLASP_OPTIONS.append('--models=0')
# ASP_CLASP_OPTIONS     = []
# ASP_CLASP_OPTIONS.append('--opt-strategy=bb,1') # default and best with frumpy + Vsids
# ASP_CLASP_OPTIONS.append('--opt-strategy=bb,2') # better in default mode
# ASP_CLASP_OPTIONS.append('--opt-strategy=bb,3') # longest of the bb
# ASP_CLASP_OPTIONS.append('--opt-strategy=usc,1') # usc seems to never finish…
# ASP_CLASP_OPTIONS.append('--opt-strategy=usc,2')
# ASP_CLASP_OPTIONS.append('--opt-strategy=usc,3')


# ASP_CLASP_OPTIONS.append('--heuristic=Berkmin') # default and ok
# ASP_CLASP_OPTIONS.append('--heuristic=Vsids') # the faster
# ASP_CLASP_OPTIONS.append('--heuristic=Vmtf')
# ASP_CLASP_OPTIONS.append('--heuristic=Domain')
# ASP_CLASP_OPTIONS.append('--heuristic=Unit')
# ASP_CLASP_OPTIONS.append('--heuristic=None')


# Definition of the default program options
PROGRAM_OPTIONS = {
    'graph_data'      : None,
    'output_file'     : None,
    'output_format'   : BUBBLE_FORMAT_ID,
    'interactive'     : False,
    'count_model'     : False,
    'count_cc'        : False,
    'timers'          : False,
    'loglevel'        : DEFAULT_LOG_LEVEL,
    'logfile'         : DEFAULT_LOG_FILE,
    'stats_file'      : None,
    'plot_stats'      : False,
    'plot_file'       : False,
    'profiling'       : False,
    'thread'          : 1,
    'draw_lattice'    : None,
}


# Functions
def basename(filepath):
    """Return the basename of given filepath.

    >>> basename('~/ASP/serious/business/fly.lp')
    'fly'

    """
    return os.path.splitext(os.path.basename(filepath))[0]

def extension(filepath):
    """Return the extension of given filepath.

    >>> extension('~/ASP/serious/business/fly.lp')
    'lp'
    >>> extension('whatisthat')
    ''
    >>> extension('whatis.that')
    'that'

    """
    return os.path.splitext(os.path.basename(filepath))[1][1:]

def is_valid_path(filepath):
    """True iff given filepath is a valid one (a file exists, or could exists)"""
    if filepath and not os.access(filepath, os.W_OK):
        try:
            open(filepath, 'w').close()
            os.unlink(filepath)
            return True
        except OSError:
            return False
    else:  # path is accessible
        return True


def test_case(filename):
    """Return path to given filename in test case, or None if the test case
    doesn't exists"""
    path = PACKAGE_DIR_TESTS + filename
    return path if os.path.exists(path) else None


def options(*, cli_doc=None, parameters={}):
    """Return the default compression options, enriched with result of CLI
    parsing if docopt documentation is given, and with parameters if given.

    All None values are put away, and the returned object is a ChainMap that
    use the default configuration for non given parameters.

    If documentation is None, CLI will not be parsed. Else docopt will be used.

    Parameters have precedence over CLI in returned options.

    """
    IRRELEVANT_CLI_OPTIONS = ('--version', '--help')
    if cli_doc:
        docopt_args = docopt(cli_doc, version=PACKAGE_VERSION)
    else:
        docopt_args = {}
    # get set of CLI parameters
    cli_args = set(
        arg.split('=')[0] if '=' in arg else arg
        for arg in sys.argv
    )
    # filter out None values
    parameters = {k: v for k, v in parameters.items() if v is not None}
    # get docopt result only for args effectively present in CLI
    docopt_args = {
        name: value
        for name, value in docopt_args.items()
        if name in cli_args
    }
    # CLI args are those returned by docopt, but corrected and filtered
    cli_args = {
        arg.lstrip('-').replace('-', '_'): value
        for arg, value in docopt_args.items()
        if value is not None and arg not in IRRELEVANT_CLI_OPTIONS
    }
    # get arguments given by parameters
    parameters = {
        arg: value
        for arg, value in parameters.items()
        if value is not None
    }
    # raise error in case of unexpected argument
    if any(arg not in PROGRAM_OPTIONS for arg in cli_args):
        unexpected_args = (arg for arg in cli_args
                           if arg not in PROGRAM_OPTIONS)
        raise ValueError(
            'ERROR: ' + str(tuple(unexpected_args)) + ' arguments is not in '
            + str(tuple(PROGRAM_OPTIONS.keys()))
        )
    return ChainMap(parameters, cli_args, PROGRAM_OPTIONS)

def thread(number=None):
    """Return Clasp options for use n thread, or if n is invalid, use the
    number of CPU available"""
    if number is not None and 1 <= int(number):
        if int(number) == 1:
            return ''
        else:
            return ' --parallel-mode=' + str(number)
    else:  # number is None, or invalid
        nb_cpu = str(multiprocessing.cpu_count())
        logger().info('Threading: Non valid number of CPU given ('
                      + str(number) + '). ' + nb_cpu
                      + ' CPU will be used by Clasp.')
        return ' --parallel-mode=' + nb_cpu


# logging functions
def logger(name=LOGGER_NAME):
    """Return logger of given name, without initialize it.

    Equivalent of logging.getLogger() call.

    """
    return logging.getLogger(name)


def configure_logger(log_filename=None, term_loglevel=None, loglevel=None):
    """Operate the logger configuration for the package"""
    # use defaults if None given
    log_filename = DEFAULT_LOG_FILE if log_filename is None else log_filename
    term_loglevel = DEFAULT_LOG_LEVEL if term_loglevel is None else term_loglevel
    loglevel = DEFAULT_LOG_LEVEL if loglevel is None else loglevel
    # put given log level in upper case, or keep it as integer
    try:
        loglevel = loglevel.upper()
    except AttributeError:  # loglevel is an integer, not a string
        pass  # nothing to do, lets keep the log level as an int
    try:
        term_loglevel = term_loglevel.upper()
    except AttributeError:  # term_loglevel is an integer, not a string
        pass  # nothing to do, lets keep the log level as an int
    assert isinstance(term_loglevel, int) or isinstance(term_loglevel, str)
    assert isinstance(loglevel, int) or isinstance(loglevel, str)

    # define the configuration
    logging_config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s',
            },
            'simple': {
                'format': '%(levelname)s %(message)s',
            },
        },
        'handlers': {
            'console': {
                'level': term_loglevel,
                'class': 'logging.StreamHandler',
                'formatter': 'simple',
            },
            'logfile': {
                'level': loglevel,
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': log_filename,
                'mode': 'w',
                'maxBytes': 2**16,
                'formatter': 'verbose',
            },
        },
        'loggers': {
            PACKAGE_NAME: {
                'handlers':['console', 'logfile'],
                'propagate': True,
                'level': loglevel,
            },
        }
    }

    # apply the configuration
    try:
        # free possible previous configuration
        handlers = logger().handlers[:]
        for handler in handlers:
            handler.close()
            logger().removeHandler(handler)
        logging.config.dictConfig(logging_config)
    except PermissionError:
        logger().warning(os.path.abspath(DIR_LOGS + LOGGER_NAME + '.log')
                        + "can't be written because of a permission error."
                        + "No logs will be saved in file.")


def log_level(level):
    """Set terminal log level to given one"""
    _logger = logger()
    handlers = (_ for _ in _logger.handlers
                if isinstance(_, logging.StreamHandler)
               )
    for handler in handlers:
        handler.setLevel(level.upper())



# -*- coding: utf-8 -*-
"""
definitions of converter base class OutConverter.

Converters are objects that convert a flow of ASP atoms
 in a particular string format.

"""
import powergrasp.commons as commons
import itertools
import re


logger = commons.logger()

REGEX_PWRNS    = re.compile(r"^powernode\(([^,]+),([^,]+),([^,]+),(.+)\)$")
REGEX_TRIVIAL  = re.compile(r"^trivial\(([^,]+),([^,]+),([^,]+)\)$")
REGEX_CLIQUE   = re.compile(r"^clique\(([^,]+),([^,]+)\)$")
REGEX_EDGES    = re.compile(r"^edge\(([^,]+),([^\)]+)\)$")
REGEX_SUBPOWER = re.compile(r"^p\(([^,]+),([^,]+),(.+)\)$")
REGEX_INC_NODE = re.compile(r"^include\(p\(([^,]+),([^,]+),([^,]+)\),([^\)]+)\)$")
REGEX_INC_PWRN = re.compile(r"^include\(p\(([^,]+),([^,]+),([^,]+)\),p\(([^,]+),([^,]+),([^,]+)\)\)$")


def multiple_matched_splits(atoms, separator, regexs):
    """Return one generator of values finds by each regex on
     each substring of given string splitted by given separator.
    """
    atoms = (a for a in str(atoms).split(separator))
    matchs = itertools.tee(atoms, len(regexs))

    matchs = ((regex.match(a)
                for a in match
              )
              for regex, match in zip(regexs, matchs)
             )
    return ((a.groups() for a in match if a is not None)
            for match in matchs
           )

def matched_splits(atoms, separator, regex):
    """Return generator of values finds by given regex on
     each substring of given string splitted by given separator.
    """
    matchs = (regex.match(a) for a in str(atoms).split(separator))
    return (a.groups() for a in matchs if a is not None)



class OutConverter:
    """Base class for Converters.

    Converters take ASP atoms (gringo.Fun), and,
     when finalizing, generate all graph description
     in the expected format.

    OutConverter is unusable as is :
     it generates ASP output.

    """
    def __init__(self):
        # for the first convertion, self.converted must be iterable
        self.atoms = tuple()

    def convert(self, atoms):
        """Operate convertion on given atoms.

        Atoms are expecting to be powernodes like:
            powernode(a,1,1,a).
        or cliques like:
            clique(a,1).
        or edges like:
            edge(a,b).

        These atoms will be used for perform the conversion
         in the finalize method.

        """
        self.atoms = itertools.chain(
            self.atoms,
            self._convert(atoms)
        )

    def _convert(self, atoms):
        """Perform the convertion and return its results

        Wait for atoms powernode(cc,k,num_set,X)
         and for clique(cc,k)
         or  for edges(X,Y).

        """
        return str(atoms)

    def header(self):
        """Return header of the file"""
        return ''

    def finalized(self):
        """Return converted data"""
        return '\n' + '\n'.join(self.atoms)

    def __str__(self):
        """str version of a converter is the finalized string of data"""
        return self.finalized()

    def comment(self, lines):
        """Add given iterable of lines as comments"""
        return '\n# ' + '\n# '.join(lines)



# -*- coding: utf-8 -*-
"""
definition of the bubble format converter.
A bubble format file (version 1), is structured in four parts:
    - list of nodes
    - list of powernodes
    - containers and contained
    - edges

There is a grammar that can be associated:
S     ::= LINE . S + FF
LINE  ::= NODES + SETS + INS + EDGES
NODES ::= 'NODE\t' . node_name      . '\n'
SETS  ::= 'SET\t'  . powernode_name . '\t' . coefficient . '\n'
INS   ::= 'IN\t'   . contained      . '\t' . container   . '\n'
EDGES ::= 'EDGE\t' . elementA       . '\t' . elementB    . '\t' . coefficient . '\n'

Coefficient are currently set to 1.0, and no tests have been performed,
so we don't know what is the purpose of it.
For contained and container, only the direct inclusion must be provided.
The order of the parts/lines can be changed.
Comments are marked by '#' character.

The Cytoscape plugin CyOoG is able to read and print that format.


"""
import itertools
from collections         import defaultdict

from powergrasp.commons  import basename, ASP_SRC_INCLUSION
from powergrasp.converter.output_converter import OutConverter
from powergrasp import info
from powergrasp import commons
from powergrasp import solving
from powergrasp import atoms


logger         = commons.logger()
GRINGO_OPTIONS = '--warn=no-atom-undefined'
GRINGO_OPTIONS = ''


def powernode(cc, step, num_set):
    """Return string representation of powernode of given

    this string representation is not ASP valid,
     but can be used as name in most file formats :

        PWRN-<cc>-<step>-<num_set>

    """
    return 'PWRN-' + '-'.join((cc, step, num_set))


def subpowernode_to_powernode(subpowernode):
    """Convert ASP representation of powernode reference (p/3)
     in powernode representation.

    Equivalent to a call to powernode(3) on the three values
     returned by REGEX_SUBPOWER.match(subpowernode).
    """
    return powernode(*REGEX_SUBPOWER.match(subpowernode).groups())


class OutBBL(OutConverter):
    """Convert given atoms in BBL format

    Used bubble format is Bubble V1.0,
     and is used by Powergraph module of BIOTEC
     for save the powernodes cytoscape representation.

    Because of the complexity of the format,
     the complexity is higher than NNFConverter.
     The convert operation stores all atoms without treatment,
      and, for big graphs, performed treatments could be heavy.
    An ASP solver is used for perform complex treatments.

    """
    META_DATA = (
        "File written by the " + info.__name__
        + " module (" + info.PACKAGE_VERSION + ")"
    )

    def __init__(self):
        super().__init__()
        self.reset_containers()

    def reset_containers(self):
        """When the data is generated, all containers can be freed
        for memory occupation optimization.

        Initialize all containers.
        """
        # nodes
        self.atoms      = tuple()
        self.nodes      = set()
        self.pwnds      = set()
        self.cliques    = set()
        # contains is a dict of container:contained
        #  it keep only one node, and its used only
        #  for powernodes with only one node inside
        self.contains   = defaultdict(set)
        # a {power,}node is in top iff is not contained by another
        self.tops       = set()
        # a {power,}node is trivial if it contains only one element
        self.trivials   = set()
        # linking between nodes
        self.edges      = defaultdict(set)
        self.belongs    = dict()

    def _convert(self, atoms):
        """Accumulate atoms"""
        return atoms

    def _additionnal_data_from(self):
        """Collect additionnal data on atoms,
        notabily on direct inclusion and triviality.

        New atoms are generated and returned.

            powernodes -- ASP source code that contains
                atoms like 'powernode(cc,k,t,x).'

        Returns many generators:
            powernode -- powernodes that are contained by nothing
            poweredge -- poweredges that links powernode and {power,}node
            clique -- powernodes that are cliques
            edge -- edges that links elements between nodes
            top -- powernodes that are contained by nothing
            topnode -- nodes that are contained by nothing
            trivial -- powernodes that contains only one element
            include_pwrn -- powernode that include a powernode
            include_node -- powernode that include a node

        """
        # create solver
        self.atoms = tuple(self.atoms)
        input_atoms = atoms.to_str(self.atoms)
        new_atoms = solving.model_from(input_atoms, aspconfig=solving.CONFIG_INCLUSION())
        if new_atoms is None:
            error = "No model generated by inclusion.lp ASP program"
            logger.critical(error)
            logger.critical(input_atoms)
            assert new_atoms is not None, error
        logger.debug('DEBUG INCLUSION: input     == ' + input_atoms)
        logger.debug('DEBUG INCLUSION: new_atoms == ' + str(new_atoms))
        new_atoms = tuple((a.predicate, a.arguments) for a in new_atoms)
        return (
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'powernode'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'poweredge'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'clique'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'edge'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'top'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'topnode'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'trivial'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'include_pwrn'),
            ((str(_) for _ in arg) for name, arg in new_atoms if name == 'include_node'),
        )

    def _generate_metadata(self):
        """assign to self metadata about knowed atoms.

        Metadata will be used by finalized method.
        """
        # get additionnal data
        (powernodes, poweredges, cliques,
         edges, tops, topnodes, trivials,
         inclusions_powernode, inclusions_node) = self._additionnal_data_from()

        # cliques cc, step (powernode cc,step is a clique)
        for cc, step in cliques:
            assert(int(step) > 0)
            self.cliques.add(powernode(cc, step, '1'))
            logger.debug('CLIQUE:' + powernode(cc, step, '1'))

        # edges a, b (there is an edge between a and b)
        for a, b in edges:
            assert(a.__class__ is str and b.__class__ is str)
            if a > b:
                a, b = b, a
            self.edges[a].add(b)
            logger.debug('EDGE:' + a + ' to ' + b)

        # edges a, b (there is an edge between a and b)
        for payload in poweredges:
            payload = tuple(payload)
            if len(payload) == 5: # powernode linked to powernode
                cc, k1, t1, k2, t2 = payload
                a = powernode(cc, k1, t1)
                b = powernode(cc, k2, t2)
            else:  # powernode linked to a node
                assert(len(payload) == 4)
                cc, k, t, node = payload
                a = powernode(cc, k, t)
                b = node
            assert(a.__class__ is str and b.__class__ is str)
            self.edges[a].add(b)
            logger.debug('POWEREDGE:' + a + ' to ' + b)

        # trivial cc, step, num_set (a powernode contains only one node)
        for cc, step, num_set in trivials:
            assert(int(step) > 0)
            assert(num_set in ('1', '2'))
            assert(powernode(cc,step,num_set) not in self.trivials)
            self.trivials.add(powernode(cc,step,num_set))
            logger.debug('TRIVIAL:' + powernode(cc,step,num_set))

        # include cc1, step1, num_set1, cc2, step2, num_set2
        #  (powernode2 is in powernode1)
        for cc1, step1, num_set1, cc2, step2, num_set2 in inclusions_powernode:
            assert(int(step1) > 0 and int(step2) > 0)
            assert(num_set1 in ('1', '2') and num_set2 in ('1', '2'))
            pwrn1 = powernode(cc1, step1, num_set1)
            pwrn2 = powernode(cc2, step2, num_set2)
            if not (pwrn2 not in self.belongs):
                logger.error("ABNORMAL SITUATION: assert(pwrn2 not in self.belongs) FAILED. WITH:")
                logger.error('PWRN1:' + pwrn1)
                logger.error('PWRN2:' + pwrn2)
                logger.error('BELONGS:' + str(self.belongs))
                # assert(pwrn2 not in self.belongs)
            self.belongs[pwrn2] = pwrn1
            logger.debug('INC_PWRN:' + pwrn1 + " contains " + pwrn2)
            self.contains[pwrn1].add(pwrn2)

        # include cc, step, num_set, node
        #  (node is in powernode cc,step,num_set)
        contained = set()
        for cc, step, num_set, node in inclusions_node:
            assert(int(step) > 0)
            assert(num_set in ('1', '2'))
            assert(node not in contained) # False iff cliques are not properly managed
            contained.add(node)
            self.nodes.add(node)
            pwrn = powernode(cc, step, num_set)
            assert(node not in self.belongs)
            self.belongs[node] = pwrn
            self.contains[pwrn].add(node)
            logger.debug('INC_NODE:' + pwrn + ' contains ' + node)

        # top cc, step, num_set (a powernode is contained by nothing)
        for node in topnodes:
            self.tops.add(node[0])
            logger.debug('TOP:' + node[0])

        # top cc, step, num_set (a powernode is contained by nothing)
        for cc, step, num_set in tops:
            self.tops.add(powernode(cc,step,num_set))
            logger.debug('TOP:' + cc + step + num_set)

    def header(self):
        return '#BBL-1.0\n#' + OutBBL.META_DATA + '\n'

    def finalized(self):
        self._generate_metadata()
        # define NODEs, SETs, INs and EDGEs relations, and return them
        return '\n' + '\n'.join(itertools.chain(
        # NODEs
            ('NODE\t' + node for node in self.nodes),
        # SETs
            # defines all powernodes
            ('SET\t' + pwnd + '\t1.0'
             for pwnd in self.pwnds
            ),
            # defines all trivial powernodes
            # NB: trivial powernodes are not accepted by
            #   powergraph cytoscape plugin. Trivials must be replaced by components
            ('SET\t' + pwnd + '\t1.0'
             for pwnd in self.trivials
            ),
        # INs
            # include in container all nodes contained
            ('IN\t'
             + (contained if contained not in self.trivials
                else next(iter(self.contains[contained]))
               )
             + '\t'
             + container
             for container, containeds in self.contains.items()
             for contained in containeds
            ),
        # EDGEs
            # create reflexives edges for cliques
            ('EDGE\t' + clique + '\t' + clique + '\t1.0'
             for clique in self.cliques
            ),
            # create edges
            ('EDGE\t' + node + '\t' + target + '\t1.0'
             for node, targets in self.edges.items()
             for target in targets
            ),
        ))



"""
definitions of converter classes.

Supported formats :
- asp: pure asp;
- nnf: nested network format;
- bbl: bubble format, used by Powergraph for printings;

"""
from powergrasp.converter.output_converter import OutConverter
from powergrasp.converter.input_converter  import InConverter
from powergrasp.converter.out_nnf          import OutNNF
from powergrasp.converter.out_bbl          import OutBBL
from powergrasp.converter.in_sbml          import InSBML
from powergrasp.converter.in_gml           import InGML
from powergrasp.converter.in_asp           import InASP
from powergrasp import commons


LOGGER = commons.logger()

# Link between format names and atom2format functions
INPUT_FORMAT_CONVERTERS = {
    ''     : InASP,  # if no extension, assume that is ASP
    'lp'   : InASP,
    'sbml' : InSBML,
    'gml'  : InGML,
}
OUTPUT_FORMAT_CONVERTERS = {
    ''    : OutBBL,  # if no extension, assume that is Bubble
    'asp' : OutConverter,
    'bbl' : OutBBL,
    'nnf' : OutBBL,
}

INPUT_FORMATS  = tuple(INPUT_FORMAT_CONVERTERS.keys())
OUTPUT_FORMATS = tuple(OUTPUT_FORMAT_CONVERTERS.keys())
DEFAULT_OUTPUT_FORMAT = 'bbl'


# converters access
def output_converter_for(format): return converter_for(format, is_output=True)
def  input_converter_for(format): return converter_for(format, is_output=False)
def delete_temporary_file(): return InConverter.delete_temporary_file()


def converter_for(format, is_output):
    """Return function that take atoms and convert them to input format or None
    """
    formats = OUTPUT_FORMAT_CONVERTERS if is_output else INPUT_FORMAT_CONVERTERS
    if format not in formats:
        LOGGER.error(
            "given extension " + str(format) + " not recognized. "
            + 'Supported ' + ('output' if is_output else 'input')
            + (' formats are ' + ', '.join(str(_) for _ in formats) + '.')
        )
        return None
    else:
        return formats[format]()


def to_asp_file(input_filename, format=None):
    """Return a dictionnary {node: succ} that contains data in input file"""
    if format is None:
        format = commons.extension(input_filename)
    return input_converter_for(format).convert(input_filename)



"""
Definition of various functions with no direct link to the package
or the compression.

"""

import os.path
import itertools
from collections import defaultdict

from powergrasp import atoms
from powergrasp import solving
from powergrasp.commons import basename


LATTICE_FILENAME   = 'powergrasp/data/lattice'  # extension is added by graphviz
ASP_FILE_INTEGRITY = 'powergrasp/ASPsources/integrity.lp'


def asp2graph(asp_atoms):
    """Convert string containing edge(X,Y) to dict {node: {succs}}"""
    graph = defaultdict(set)
    print(asp_atoms)
    edges = (atom for atom in asp_atoms.split('.') if atom.startswith('inter('))
    for atom, args in (atoms.split(atom) for atom in edges):
        x, y = args
        graph[x].add(y.upper())
        graph[y].add(x.upper())
    return graph

def line_diagram(graph, filename=LATTICE_FILENAME):
    import concepts
    diag = concepts.Definition()
    for node in sorted(graph.keys()):
        diag.add_object(node, graph[node])
    context = concepts.Context(*diag)
    context.lattice.graphviz().render(filename=os.path.basename(filename),
                                      directory=os.path.dirname(filename),
                                      cleanup=True)  # cleanup the dot file
    assert os.path.exists(filename + '.pdf')  # output is pdf


def test_integrity(asp_graph_data_filename,
                   asp_file_integrity=ASP_FILE_INTEGRITY):
    """Perform integrity and consistency study of the graph described
    by ASP edge/2 atoms contained in file.

    Return a dict property:value.
    """
    with open(asp_graph_data_filename) as fd:
        graph_atoms = ''.join(l for l in fd.read() if l not in '\n ')
    model = solving.model_from(graph_atoms, ASP_FILE_INTEGRITY)
    assert model is not None
    return {
        name + ' ' + ': '.join(str(_) for _ in args)
        for name, args in (atoms.split(atom) for atom in model)
    }


def dict2atoms(graph, converted_graph_filename):
    """write in given filename the equivalent to given graph in ASP"""
    with open(converted_graph_filename, 'w') as fd:
        for node, targets in converted_graph_filename.items():
            fd.write('\n'.join(
                'edge("'
                + str(node)
                + '","' + str(n)
                + '").'
                for n in targets
            ))

def make_clique(nb_node, filename):
    """write in given file a graph that is a clique of nb_node node"""
    def int2node(i) : return '"n' + str(i+1) + '"'
    with open(filename, 'w') as fd:
            [fd.write('edge(' + int2node(idx) + ',' + int2node(linked) + ').\n')
             for idx in range(nb_node)
             for linked in range(idx+1, nb_node)
            ]


def make_tree(nb_node, filename, nb_child=lambda:2):
    """write in given file a graph that is a bintree of nb_node node"""
    import math
    carac   = (chr(_) for _ in range(ord('a'), ord('z')+1))
    nb_cars = int(math.ceil(math.log(nb_node, 26)))
    label   = (''.join(c)
               for c in itertools.islice(itertools.product(
                carac, repeat=nb_cars), nb_node
               ))
    def childs():
        for _ in range(nb_child()):
            yield next(label)
    def edge(n, m):
        return 'edge(' + n + ',' + m + ').\n'
    nodes = [next(label)]
    with open(filename, 'w') as fd:
        for node in nodes:
            for child in childs():
                fd.write(edge(node, child))
                nodes.append(child)







"""
definition of the GML input format converter.

"""
from powergrasp.converter.input_converter import InConverter
import powergrasp.commons as commons
import itertools


LOGGER = commons.logger()


class InGML(InConverter):
    """Convert given GML file in uniformized data"""
    FORMAT_NAME = 'gml'
    FORMAT_EXTENSIONS = ('gml',)

    def _gen_edges(self, filename_gml):
        """Yields pair (node, successor), representing the data contained
        in input gml file.
        """
        try:
            from networkx import read_gml
            graph = read_gml(filename_gml)
            for node1, node2 in graph.edges():
                yield node1, node2
        except IOError as e:
            LOGGER.error(self.error_input_file(filename_sbml, e))
            exit(1)
        except ImportError:
            LOGGER.critical('networkx python module is necessary to'
                            ' use GML as input format')
            exit(1)



"""
Main source file of the package, containing tho compress function.

The compress function get numerous arguments,
 for allowing a parametrable compression.

"""
import os
import inspect
import tempfile
from builtins           import input
from collections        import defaultdict

from powergrasp.observers import Signals  # shortcut
from powergrasp.commons import basename
from powergrasp.commons import ASP_SRC_EXTRACT, ASP_SRC_PREPRO , ASP_SRC_FINDCC
from powergrasp.commons import ASP_SRC_FINDBC , ASP_SRC_POSTPRO, ASP_SRC_POSTPRO
from powergrasp.commons import ASP_ARG_UPPERBOUND, ASP_ARG_CC
from powergrasp.commons import ASP_ARG_LOWERBOUND, ASP_ARG_STEP
from powergrasp import compression
from powergrasp import statistics
from powergrasp import observers
from powergrasp import converter
from powergrasp import solving
from powergrasp import commons
from powergrasp import atoms



LOGGER = commons.logger()


def network_name_from(data):
    """A string describing the graph data received.

    if data is a valid filepath, the filepath will be returned.
    Else, the string 'network' will be returned.

    """
    if os.path.isfile(data):
        return data
    else:
        return 'stdin network'


def asp_file_from(data):
    """A filename containing the graph data formatted in ASP.

    Data is a filename, or a string containing the graph data,
    encoded in ASP format.
    Detection of the type of data (filename or graph) is performed by detect
     a valid path in data. On failure, data is understood as an input graph.

    """
    # default case: data is a filename
    graph_data_file = data
    data_format = None
    # data is a string formatted in an input format.
    # try to access data
    if not commons.is_valid_path(data):
        LOGGER.info('Input data is not a valid path. This data is assumed as'
                    ' ASP formatted data.')
        file_to_be_converted = tempfile.NamedTemporaryFile('w', delete=False)
        file_to_be_converted.write(data)
        file_to_be_converted.close()
        graph_data_file = file_to_be_converted.name
        data_format = 'asp'
    elif not os.path.exists(data):
        # the file is not existing, raise the error !
        open(data)
    # convert graph data into ASP-readable format
    graph = converter.to_asp_file(graph_data_file, format=data_format)
    return graph_dict_to_asp_file(graph)


def graph_dict_to_asp_file(graph_dict):
    """convert {node: succs} to ASP atoms edge/2, where edge(X,Y) defines X
    as node and Y a successor.

    Returns the temp file name where atoms are pushed.

    """
    # write it in a file, and convert this file in ASP.
    asp_file = tempfile.NamedTemporaryFile('w', delete=False)
    def to_asp_value(value):
        if isinstance(value, int):
            return str(value)
        return (  # surround value if necessary
            ('"' if value[0] != '"' else '')
            + str(value)
            + ('"' if value[-1] != '"' else '')
        )
    for node, succs in graph_dict.items():
        for succ in succs:
            asp_file.write('edge(' + to_asp_value(node) + ','
                           + to_asp_value(succ) + ').\n')
    asp_file.close()
    return asp_file.name


def compress(graph_data=None, output_file=None, *,
             output_format=None, interactive=None,
             count_model=None, count_cc=None,
             stats_file=None, timers=None, logfile=None, loglevel=None,
             thread=None, draw_lattice=None, instanciated_observers=None,
             extract_config=None, biclique_config=None, clique_config=None):
    """Performs the graph compression with data found in graph file.

    Any not given argument will be overriden by default values.

    Output format must be a valid string,
     or will be inferred from the output file name, or will be set as bbl.

    If output file is None, result will be printed in stdout.

    The function itself returns a float that is, in seconds,
     the time necessary for the compression,
     and the object provided by the statistics module,
     that contains statistics about the compression.

    """
    # define the log file and the log level, if necessary
    commons.configure_logger(logfile, loglevel)

    # None to default
    if extract_config is None:
        extract_config = solving.CONFIG_EXTRACTION()
    if biclique_config is None:
        biclique_config = solving.CONFIG_BICLIQUE_SEARCH()
    if clique_config is None:
        clique_config = solving.CONFIG_CLIQUE_SEARCH()

    # gives default value for each parameter that needs it
    _, _, _, func_args = inspect.getargvalues(inspect.currentframe())
    func_args = dict(func_args)  # copy data structure
    option = commons.options(parameters=func_args)

    # configs enrichment
    thread_option = commons.thread(option['thread'])
    if thread_option:
        extract_config = solving.ASPConfig(extract_config.files,
                                           extract_config.clasp_options + thread_option,
                                           extract_config.gringo_options)
        clique_config = solving.ASPConfig(clique_config.files,
                                          clique_config.clasp_options + thread_option,
                                           clique_config.gringo_options)
        biclique_config = solving.ASPConfig(biclique_config.files,
                                            biclique_config.clasp_options + thread_option,
                                            biclique_config.gringo_options)

    # get data from parameters
    graph_file = asp_file_from(option['graph_data'])
    # Create the default observers
    output_converter = observers.OutputWriter(option['output_file'],
                                              option['output_format'])
    if instanciated_observers is None:  # default value handling
        instanciated_observers = []
    instanciated_observers += [
        output_converter,
    ]
    # Add the optional observers
    if option['timers']:
        time_counter = observers.TimeCounter(ignore=[
            Signals.IterationStarted, Signals.PreprocessingStarted,
        ])
    else:  # no timers asked, but others modules may want to have a ref to
        time_counter = observers.NullTimeCounter()
    instanciated_observers.append(time_counter)

    if option['stats_file']:
        instanciated_observers.append(statistics.DataExtractor(
            stats_file,
            output_converter=output_converter,
            time_counter=time_counter,
            network_name=network_name_from(graph_data)
        ))

    if option['count_model']:
        instanciated_observers.append(observers.ObjectCounter())
    if option['count_cc']:
        instanciated_observers.append(observers.ConnectedComponentsCounter())

    if option['draw_lattice']:
        instanciated_observers.append(observers.LatticeDrawer(draw_lattice))
    if option['interactive']:
        instanciated_observers.append(observers.InteractiveCompression())

    # sort observers, in respect of their priority (smaller is after)
    instanciated_observers.sort(key=lambda o: o.priority.value, reverse=True)
    assert instanciated_observers[0].priority.value >= instanciated_observers[-1].priority.value
    LOGGER.debug('OBSERVERS:' + str('\n\t'.join(
        str((obs.__class__, obs))
        for obs in instanciated_observers
    )))

    # Launch the compression
    LOGGER.info('COMPRESSION STARTED !')
    compression.compress_lp_graph(
        graph_file,
        all_observers=tuple(instanciated_observers),
        extract_config=extract_config,
        clique_config=clique_config,
        biclique_config=biclique_config,
    )
    LOGGER.info('COMPRESSION FINISHED !')



# -*- coding: utf-8 -*-
"""
definitions of converter base class InConverter.

Converters are objects that convert a flow of ASP atoms
 in a particular string format.

Supported formats :
- asp: pure asp;
- nnf: nested network format;
- bbl: bubble format, used by Powergraph for printings;

"""
import itertools
import tempfile
import re
import os
from collections import defaultdict

import powergrasp.commons as commons
from powergrasp.commons import basename


LOGGER = commons.logger()


class InConverter(object):
    """Base class for Converters.

    Converters take an input format, and convert it in internal representation
    of the graph.
    InConverter returns empty graph by default.

    Subclasses should:
        define a class constant FORMAT_NAME giving the format name
        define a class constant FORMAT_EXTENSIONS giving an iterable of file
            extensions readable by the subclass.
        redefine the _gen_edges method.

    """
    FORMAT_NAME = 'asp'
    FORMAT_EXTENSIONS = ()

    def convert(self, filename:str) -> dict:
        """Return a dict {node: successors}, representing the data contained
        in input file.
        """
        try:
            graph = defaultdict(set)
            for node, succ in self._gen_edges(filename):
                graph[node].add(succ)
            return dict(graph)
        except (IOError, PermissionError):
            LOGGER.critical('File ' + outputfile + " can't be opened."
                            + ' Graph data retrieving needs this file.'
                            + ' Compression aborted')
            return {}

    def _gen_edges(self, input_file:str) -> dict:
        """Yields pair (node, successor), representing the data contained
        in input file.

        The input file should have an extension defined
        in self.__class__.FORMAT_EXTENSIONS.

        This method is the only one that needs to be overrided
        by subclasses.

        """
        LOGGER.error('The InConverter._convert_to_dict method was called.'
                     ' Subclasses of InConverter should redefine this method.'
                     ' No data generated.')
        return  # empty generator pattern
        yield

    @classmethod
    def error_input_file(cls, filename:str, exception:Exception=None) -> str:
        """return string error for not openable input file"""
        print(exception, dir(exception))
        exit()
        return ('File ' + filename + ' can\'t be opened as an '
                + self.FORMAT_NAME + ' file.'
                + ' Compression aborted')



"""
Definition of the plotting method.

"""
from powergrasp import commons
from powergrasp import statistics

# Logger
LOGGER = commons.logger()

try:
    # plotting libraries
    #  this is just a test for print a warning if there are not present
    import matplotlib
    import pandas
    import numpy
except ImportError:
    LOGGER.warning('plotting libraries are not all there. '
                   'Maybe you will need to install them.')


# Data for plotting
MEASURES = (statistics.GENR_TIME, statistics.COMP_EDGE,
            statistics.GENR_PWED, statistics.GENR_PWND)
COLORS   = ('black', 'green', 'blue', 'red')
LABELS   = (
    'time per step',
    'edge compressed',
    'generated poweredge',
    'generated powernode',
)


# PLOTTING
def plots(filename, title="Compression statistics", xlabel='Iterations',
          ylabel='{Power,} {node,edge}s counters', savefile=None, dpi=400):
    """Generate the plot that show all the data generated by the compression

    if savefile is not None and is a filename, the figure will be saved
    in png in given file, with given dpi."""
    try:
        # plotting libraries
        from matplotlib import rc
        rc('text', usetex=True)
        import matplotlib.pyplot as plt
        import numpy as np
        import pandas as pd
        from matplotlib.pyplot import savefig
    except ImportError:
        LOGGER.error('plotting libraries are not all there. '
                     'Please install matplotlib and pandas modules. '
                     'Plotting aborted.')
        return  # end of plotting

    # GET DATA
    try:
        data = np.genfromtxt(
            filename,
            delimiter=',',
            skip_header=0,
            skip_footer=0,
            names=True,  # read names from header
        )
    except IOError as e:
        LOGGER.warning('The file '
                       + statistics_filename
                       + ' can\'t be opened. No statistics will be saved.')


    # Label conversion
    def key2label(key):
        """Convert given string in label printable by matplotlib"""
        return '$\#$' + key.replace('count', '').replace('_', ' ')

    # PLOTTING
    try:
        data_size = len(data[MEASURES[0]])
    except TypeError:
        LOGGER.error('Plotting compression statistics require'
                     ' more than one compression iteration')
        LOGGER.error('Plotting aborted')
        return

    # convert in pandas data frame for allow plotting
    gx = pd.DataFrame(data, columns=MEASURES)
    # {black dotted,red,yellow,blue} line with marker o
    styles = ['ko--', 'ro-', 'yo-', 'bo-']

    # get plot, and sets the labels for the axis and the right axis (time)
    plot = gx.plot(style=styles, secondary_y=[statistics.GENR_TIME])
    lines, labels = plot.get_legend_handles_labels()
    rines, rabels = plot.right_ax.get_legend_handles_labels()
    labels = [key2label(l) for l in labels] + ['concept generation time']

    plot.legend(lines + rines, labels)
    plot.set_xlabel(xlabel)
    plot.set_ylabel(ylabel)
    plot.right_ax.set_ylabel('Time (s)')

    # axis limits : show the 0
    plot.right_ax.set_ylim(0, max(gx[statistics.GENR_TIME]) * 2)
    plot.set_ylim(0, max(gx[statistics.COMP_EDGE]) * 1.1)

    # print or save
    if savefile:
        plt.savefig(savefile, dpi=dpi)
        LOGGER.info('Plot of statistics data saved in file '
                    + savefile + ' (' + str(dpi) + ' dpi)')
    else:
        plt.show()



# -*- coding: utf-8 -*-
"""
Definitions of many utils functions about gringo atoms manipulation.
Provides converters, access and printings of atoms.

"""

from collections        import defaultdict, Counter, namedtuple
import itertools

from pyasp import asp


# ASP Parser: if atoms are defined as arguments, consider them strings.
PARSER = asp.Parser(collapseTerms=True, collapseAtoms=False)

# Atom definition
ATOM = namedtuple('Atom', ['name', 'args'])

# CONSTANTS
RESULTS_PREDICATS = (
    'powernode',
    'poweredge',
    'score',
)


def split(atom):
    """Return the splitted version of given atom.

    atom -- string formatted as an ASP readable atom.
    return -- None or an ATOM object with field name and args.

    If many atoms are defined in the input string,
        only one will be returned.
    If given atom is not valid, None is returned.

    >>> split('edge(lacA,lacZ)')
    Atom(name='edge', args=('lacA', 'lacZ'))
    >>> split('edge(42,12)')
    Atom(name='edge', args=('42', '12'))
    >>> split('edge("ASX38","MER(HUMAN)")')
    Atom(name='edge', args=('"ASX38"', '"MER(HUMAN)"'))
    >>> split('lowerbound.')
    Atom(name='lowerbound', args=())
    >>> split('')

    """
    try:
        parsed = next(iter(PARSER.parse(atom.rstrip('.'))))
        return ATOM(parsed.predicate, tuple(parsed.args()))
    except StopIteration:
        return None


def arg(atom):
    """Return the argument of given atom, as a tuple

    >>> arg('edge(lacA,lacZ)')
    ('lacA', 'lacZ')
    >>> arg('score(13)')
    ('13',)
    >>> arg('lowerbound')
    ()

    """
    payload = atom.strip('.').strip(')')
    try:
        return tuple(payload.split('(')[1].split(','))
    except IndexError:  # no args !
        return tuple()


def first_arg(atom):
    """Return the first argument of given atom, or None if no arg

    >>> first_arg('edge(lacA,lacZ)')
    'lacA'
    >>> first_arg('score(13)')
    '13'
    >>> first_arg('lowerbound')

    """
    try:
        return arg(atom)[0]
    except IndexError:
        return None


def prettified(atoms, names=None, sizes=None,
               joiner='\n', results_only=False, sort=False, hashs=False):
    """Return a human readable string version or
      string generator of given atoms.

    if names is given, only atoms with given name will be used.
    if results_only is True, only some atoms will be printed.
      (defined by RESULTS_PREDICATS)
    if sizes is provided, atoms with args number not in sizes
      will not be printed.

    if sort is True, a non-lazy treatment will be applied to all data,
      and the atoms will be returned in sorted order.

    if hash is True, each atom will be printed with its hash.

    atoms must be an iterable of gringo.Fun instances.

    """
    atoms, hashs_values = itertools.tee(atoms)
    atoms = ((atom.predicate, atom.arguments) for atom in atoms)
    if hashs:
        hashs_values = (a.__hash__() for a in hashs_values)
    # filter results
    if results_only:
        source = ((n,a)
                  for n,a in atoms
                  if n in RESULTS_PREDICATS
                 )
    else:
        source = (_ for _ in atoms)
    # filter names
    if names is not None:
        source = ((name,args)
                  for name,args in source
                  if name in names
                 )
    # filter sizes
    if sizes is not None:
        source = (
                  (name,args)
                  for name, argset in source
                  if len(args) in sizes
                 )
    # get final text generator
    source = (
        name+'('+','.join(str(_) for _ in args)+').'
        for name, args in source
    )
    # hashs
    if hashs:
        source = (
            str(hash) + ' ' + name
            for name, hash in zip(source, hashs_values)
        )
    # sorting
    if sort:
        source = sorted(tuple(source))

    # joining if possible
    return source if joiner is None else joiner.join(source)


def count(atoms, names=None) -> Counter:
    """Return a dict atom:count that describes
    how many atoms given atoms_dict have.

    if names is None, all atoms will be returned.
    if names is an iterable of atoms names,
     only founded atoms will be returned.
    """
    if names is None:
        return Counter(atom.predicate for atom in atoms)

    if isinstance(names, str):
        names = set([names])
    return Counter(atom.predicate for atom in atoms if atom.predicate in names)


def to_str(atoms, names=None, separator='.'):
    """Return string that is equivalent and ASP-valid from
    given atoms.

    If names is provided, only atoms with given names will be returned.
     names can be a single name or a container of names.
    separator is the string that will be added between each
     and at the end of the atoms.

    """
    if names:
        atoms = atoms.get(names)
    return separator.join(atom.predicate + '(' + ','.join(atom.arguments) + ')'
                          for atom in atoms) + separator



"""
Definitions of model_from(5) that encapsulate
 the ASP grounder and solver calls.

"""
import os
import itertools
from functools   import partial
from collections import deque, Counter, namedtuple

from powergrasp import commons
from powergrasp import atoms
from pyasp import asp


LOGGER = commons.logger()


# ASP Configuration class
ASPConfig = namedtuple('ASPConfig', 'files clasp_options gringo_options')
ASPConfig.__new__.__defaults__ = None, '', ''  # constructor default values

# shortcuts for solvers options
HEURISTICS = ('Berkmin', 'Vmtf', 'Vsids', 'Unit', 'None', 'Domain')
CONFIGURATIONS = ('frumpy', 'jumpy', 'tweety', 'trendy', 'crafty', 'handy')
STRATEGIES = ('bb,1', 'bb,2', 'bb,3', 'usc,1', 'usc,2', 'usc,3')
FLAGS = ('', '--restart-on-model')

def gen_configs():
    return itertools.product(HEURISTICS, CONFIGURATIONS, STRATEGIES, FLAGS)
NB_CONFIG = len(tuple(gen_configs()))

ASP_CONF_OPTIONS = ' --heuristic={} --configuration={} --opt-strategy={} {} -n 0'
ASP_DEFAULT_CLASP_OPTION = ' --heuristic=Vsids --configuration=frumpy '

def gen_extract_configs():
    for heur, conf, strat, flag in gen_configs():
        yield ASPConfig([commons.ASP_SRC_EXTRACT], ASP_CONF_OPTIONS.format(heur, conf, strat, flag))

# default configs
CONFIG_DEFAULT = lambda: ASPConfig([])
CONFIG_EXTRACTION = lambda: ASPConfig(
    [commons.ASP_SRC_EXTRACT],
    ' --heuristic=Vsids --configuration=frumpy -n 0',
)
CONFIG_BICLIQUE_SEARCH = lambda: ASPConfig(
    [commons.ASP_SRC_PREPRO, commons.ASP_SRC_POSTPRO, commons.ASP_SRC_FINDBC],
    ASP_DEFAULT_CLASP_OPTION,
)
CONFIG_CLIQUE_SEARCH = lambda: ASPConfig(
    [commons.ASP_SRC_PREPRO, commons.ASP_SRC_POSTPRO, commons.ASP_SRC_FINDCC],
    ASP_DEFAULT_CLASP_OPTION,
)
CONFIG_INCLUSION = lambda: ASPConfig(
    [commons.ASP_SRC_INCLUSION],
)


class Atoms(asp.TermSet):
    """This class is a patching of TermSet, and is returned in place of it
    by the functions of this module."""

    def get(self, atom_names):
        if isinstance(atom_names, str):
            return (atom for atom in self if atom.predicate == atom_names)
        else:
            return (atom for atom in self if atom.predicate in atom_names)

    def get_first(self, atom_name):
        return next(atom for atom in self if atom.predicate == atom_name)

    def counter(self):
        if not hasattr(self, '_counter'):
            self._counter = Counter(atom.predicate for atom in self)
        return self._counter

    def count(self, name:str):
        if not hasattr(self, '_counter'):
            self.counter()
        return self._counter[name]


def all_models_from(base_atoms, aspfiles=None, aspargs=None,
                    aspconfig=None, parsed=True):
    """Compute all models from ASP source code in aspfiles, with aspargs
    given as grounding arguments and base_atoms given as input atoms.

    base_atoms -- string, ASP-readable atoms
    aspfiles -- (list of) filename, contains the ASP source code
    aspargs -- dict of constant:value that will be set as constants in aspfiles
    aspconfig -- an ASPConfig object giving more files and solvers options
    parsed -- set to True for get already parsed atoms of model, False for str
    yield -- Atoms instance, containing atoms produced by solving

    """
    if aspfiles is None:
        aspfiles = []
    if aspargs is None:
        aspargs = {}
    if aspconfig is None:
        aspconfig = CONFIG_DEFAULT()
    # use list of aspfiles in all cases
    assert aspfiles.__class__ in (tuple, list, str)
    assert aspconfig.__class__ is ASPConfig
    if isinstance(aspfiles, str):
        aspfiles = [aspfiles]
    elif isinstance(aspfiles, tuple):  # solver take only list, not tuples
        aspfiles = list(aspfiles)
    if aspconfig.files is not None:
        aspfiles.extend(aspconfig.files)
    assert aspfiles.__class__ is list

    # define the command line options for gringo and clasp
    constants = ' -c '.join(str(k)+'='+str(v) for k,v in aspargs.items())
    if len(aspargs) > 0:  # must begin by a -c for announce the first constant
        constants = '-c ' + constants
    gringo_options = constants + ' ' + aspconfig.gringo_options
    # print('OPTIONS:', gringo_options)
    # print('OPTIONS:', aspconfig.clasp_options)
    # print('OPTIONS:', aspconfig)

    #  create solver and ground base and program in a single ground call.
    solver = asp.Gringo4Clasp(gringo_options=gringo_options,
                              clasp_options=aspconfig.clasp_options)
    # print('SOLVING:', aspfiles, constants)
    # print('INPUT:', base_atoms.__class__, base_atoms)
    answers = solver.run(aspfiles, additionalProgramText=base_atoms,
                         collapseAtoms=not parsed)
    # if len(answers) > 0:
        # for idx, answer in enumerate(answers):
            # print('ANSWER ' + str(idx) + ':', answer)
            # print('ATOM(S):', atoms.count(answer))
    # else:
        # print('NO MODEL FOUND !')
    yield from (Atoms(answer) for answer in answers)


def model_from(base_atoms, aspfiles=None, aspargs=None, aspconfig=None,
               parsed=True):
    """Compute the last model from ASP source code in aspfiles, with aspargs
    given as grounding arguments and base_atoms given as input atoms.

    base_atoms -- string, ASP-readable atoms
    aspfiles -- (list of) filename, contains the ASP source code
    aspargs -- dict of constant:value that will be set as constants in aspfiles
    aspconfig -- an ASPConfig object giving more files and solvers options
    parsed -- set to True for get already parsed atoms of model, False for str
    return -- an Atoms instance containing atoms of the last produced model

    """
    answers = all_models_from(base_atoms=base_atoms, aspfiles=aspfiles,
                              aspargs=aspargs, aspconfig=aspconfig,
                              parsed=parsed)
    try:
        return deque(answers, maxlen=1)[0]
    except IndexError:  # no valid model
        return None



"""
Some meta informations about the package.

"""


PROGRAM_NAME    = 'PowerGrASP'
PACKAGE_NAME    = PROGRAM_NAME.lower()
__fix           = '1'
__minor         = '4'
__major         = '0'
PACKAGE_VERSION = '.'.join((__major, __minor, __fix))


# link between filenames and studies
from collections import defaultdict
STUDYNAMES = defaultdict(lambda: 'No related study')
STUDYNAMES.update({
    'proteome_yeast_1'   : 'Casein Kinase II Complex',
    'proteome_yeast_2'   : 'Histone interaction',
    'structural_binding' : 'Interactions of SH3 Carrying Proteins',
    'phosphatase'        : 'Human Protein Tyrosine Phosphatase Homology Network',
})



# -*- coding: utf-8 -*-
"""
definition of the NNF format converter.

"""
from powergrasp.converter.output_converter import OutConverter
import powergrasp.commons as commons
import itertools


logger = commons.logger()

class OutNNF(OutConverter):
    """Convert given atoms in NNF format"""

    def _convert(self, powernodes, cliques):
        """Operate convertion on given atoms"""
        # get first item for obtain global data
        cc, k, s, node = next(atoms)
        atoms = itertools.chain( ((cc,k,s,node),), atoms )

        # generate lines
        nnf = ('{0}_{1}_{2}\t{3}'.format(*g) for g in atoms)
        return itertools.chain(
            nnf,
            ('{0}_{1}\t{2}_{3}_{4}\tpp\t{5}_{6}_{7}'.format(cc,k,cc,k,1,cc,k,2),),
            ('{0}_cc\t{1}_{2}'.format(cc,cc,k),),
        )

    def _convert_edge(self, atoms):
        """Perform the convertion and return its results"""
        # get first item for obtain global data
        cc, k, s = next(atoms)
        atoms = itertools.chain( ((cc,k,s),), atoms )

        # generate lines
        nnf = ('{0}\tpp\t{2}'.format(*g) for g in atoms)
        return nnf




"""
Definition of various counting related compression observers.

"""
from .observer import CompressionObserver, Signals, SIGNAL_FOUND
from powergrasp import commons


LOGGER = commons.logger()


class ConnectedComponentsCounter(CompressionObserver):
    """Counter of connected components"""

    def __init__(self):
        self._nb_ccs = '?'  # number of cc, printed after current cc number

    def _update(self, signals):
        if Signals.CCCountGenerated in signals:
            self._nb_ccs = str(signals[Signals.CCCountGenerated])
        if Signals.CCRemainEdgeGenerated in signals:
            count = len(signals[Signals.CCRemainEdgeGenerated])
            if count > 0:
                LOGGER.info(self.cc_name + ': ' + str(count)
                            + ' remaining edge(s)')
            else:
                LOGGER.info(self.cc_name + ': ' + 'No remaining edge')
        if Signals.ConnectedComponentStarted in signals:
            self.cc_num, self.cc_name = signals[Signals.ConnectedComponentStarted]
            LOGGER.info('#### CC ' + self.cc_name + ' ' + str(self.cc_num+1)
                        + '/' + self._nb_ccs)


class ObjectCounter(CompressionObserver):
    """Counter of objects, based on received signals ending with '_found'"""

    PREFIX = '_counter_'

    def __init__(self):
        self.props = []
        for signal in (s for s in Signals if s.value.endswith(SIGNAL_FOUND)):
            prop = ObjectCounter.property_named_from(signal.value)
            setattr(self, prop, 0)
            self.props.append(prop)

    def _update(self, signals):
        for signal in signals:
            name = signal.value
            if name.endswith(SIGNAL_FOUND):
                prop = ObjectCounter.property_named_from(name)
                setattr(self, prop, getattr(self, prop) + 1)
        if Signals.CompressionFinalized in signals:
            for prop in self.props:
                LOGGER.info(self.__class__.__name__ + ': '
                            + ObjectCounter.prettified(prop)
                            + ': ' + str(getattr(self, prop)))

    @staticmethod
    def property_named_from(signal_name):
        "Return the property name deduced from the signal name"
        return ObjectCounter.PREFIX + signal_name[:-len(SIGNAL_FOUND)]

    @staticmethod
    def prettified(property_name):
        "Return the prettified property name deduced from the property name"
        property_name = property_name[len(ObjectCounter.PREFIX):]
        return property_name.replace('_', ' ').title()
